from typing import Optional, Tuple, Dict, List
from dataclasses import dataclass, field
from aas_test_engines.result import AasTestResult, Level
from enum import Enum
import datetime

from .model import Environment
from .parse_submodel import parse_submodel, LangString
from .parse import check_constraints, CheckConstraintException
from .adapter import AdapterPath

templates = {}


def template(semantic_id: str):
    assert semantic_id not in templates

    def decorator(fn):
        templates[semantic_id] = fn
        return fn
    return decorator


class RoleOfContactPerson(Enum):
    Administrative = "0173-1#07-AAS927#001"
    Commercial = "0173-1#07-AAS928#001"
    Other = "0173-1#07-AAS929#001"
    HazardousGoods = "0173-1#07-AAS930#001"
    Technical = "0173-1#07-AAS931#001"


class TypeOfTelephone(Enum):
    Office = "0173-1#07-AAS754#001"
    OfficeMobile = "0173-1#07-AAS755#001"
    Secretary = "0173-1#07-AAS756#001"
    Substitute = "0173-1#07-AAS757#001"
    Home = "0173-1#07-AAS758#001"
    PrivateMobile = "0173-1#07-AAS759#001"


@dataclass
class Phone:
    telephone_number: LangString = field(metadata={
        "semantic_id": "0173-1#02-AAO136#002",
    })
    type_of_telephone: Optional[TypeOfTelephone] = field(metadata={
        "semantic_id": "0173-1#02-AAO137#003"
    })
    available_time: Optional[LangString] = field(metadata={
        "semantic_id": "https://admin-shell.io/zvei/nameplate/1/0/ContactInformations/ContactInformation/AvailableTime/",
    })


class TypeOfFaxNumber(Enum):
    Office = "0173-1#07-AAS754#001"
    Secretary = "0173-1#07-AAS756#001"
    Home = "0173-1#07-AAS758#001"


@dataclass
class Fax:
    fax_number: LangString = field(metadata={
        "semantic_id": "0173-1#02-AAO195#002",
    })
    type_of_fax_number: Optional[TypeOfFaxNumber] = field(metadata={
        "semantic_id": "0173-1#02-AAO196#003",
    })


class TypeOfEmailAddress(Enum):
    Office = "0173-1#07-AAS754#001"
    Secretary = "0173-1#07-AAS756#001"
    Substitute = "0173-1#07-AAS757#001"
    Home = "0173-1#07-AAS758#001"


@dataclass
class Email:
    email_address: str = field(metadata={
        "semantic_id": "0173-1#02-AAO198#002",
    })
    public_key: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO200#002",
    })
    type_of_email_address: Optional[TypeOfEmailAddress] = field(metadata={
        "semantic_id": "0173-1#02-AAO199#003"
    })
    type_of_public_key: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO201#002"
    })


@dataclass
class IPCommunication:
    address_of_additional_link: str = field(metadata={
        "semantic_id": "0173-1#02-AAQ326#002",
    })
    type_of_communication: Optional[str] = field(metadata={
        "semantic_id": "https://admin-shell.io/zvei/nameplate/1/0/ContactInformations/ContactInformation/IPCommunication/TypeOfCommunication",
    })
    available_time: Optional[str] = field(metadata={
        "semantic_id": "https://admin-shell.io/zvei/nameplate/1/0/ContactInformations/ContactInformation/AvailableTime/",
    })


@dataclass
class ContactInformation:
    role_of_contact_person: Optional[RoleOfContactPerson] = field(metadata={
        "semantic_id": "0173-1#02-AAO204#003",
    })
    national_code: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO134#002",
    })
    language: Optional[List[str]] = field(metadata={
        "semantic_id": "https://admin-shell.io/zvei/nameplate/1/0/ContactInformations/ContactInformation/Language",
    })
    time_zone: Optional[str] = field(metadata={
        "semantic_id": "https://admin-shell.io/zvei/nameplate/1/0/ContactInformations/ContactInformation/TimeZone",
    })
    # TODO: Add hint: "mandatory property according to EU MachineDirective 2006/42/EC."
    city_town: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO132#002",
    })
    company: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAW001#001",
    })
    department: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO127#003",
    })
    phone: Optional[Phone] = field(metadata={
        "semantic_id": "https://admin-shell.io/zvei/nameplate/1/0/ContactInformations/ContactInformation/Phone",
    })
    fax: Optional[Fax] = field(metadata={
        "semantic_id": "0173-1#02-AAQ834#005",
    })
    email: Optional[Email] = field(metadata={
        "semantic_id": "0173-1#02-AAQ836#005",
    })
    ipc_communication: Optional[List[IPCommunication]] = field(metadata={
        "semantic_id": "https://admin-shell.io/zvei/nameplate/1/0/ContactInformations/ContactInformation/IPCommunication",
    })
    street: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO128#002",
    })
    zip_code: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO129#002",
    })
    po_box: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO130#002",
    })
    zip_code_of_po_box: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO131#002",
    })
    state_country: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO133#002",
    })
    name_of_contact: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO205#002",
    })
    first_name: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO206#002",
    })
    middle_name: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO207#002",
    })
    title: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO208#003",
    })
    academic_title: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO209#003",
    })
    further_details_of_contact: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO210#002",
    })
    address_of_additional_link: Optional[str] = field(metadata={
        "semantic_id": "] 0173-1#02-AAQ326#002",
    })


@dataclass
@template("https://admin-shell.io/zvei/nameplate/1/0/ContactInformations")
class ContactInformations:
    contact_information: List[ContactInformation] = field(metadata={
        "semantic_id": "https://admin-shell.io/zvei/nameplate/1/0/ContactInformations/ContactInformation",
    })


@dataclass
@template("https://admin-shell.io/zvei/nameplate/2/0/Nameplate")
class DigitalNameplate:
    uri_of_the_product: str = field(metadata={
        'semantic_id': '0173-1#02-AAY811#001',
    })
    manufacturer_name: LangString = field(metadata={
        'semantic_id': '0173-1#02-AAO677#002',
    })
    manufacturer_product_designation: LangString = field(metadata={
        'semantic_id': '0173-1#02-AAW338#001',
    })
    contact_information: ContactInformation = field(metadata={
        "semantic_id": "https://admin-shell.io/zvei/nameplate/1/0/ContactInformations/ContactInformation",
    })
    manufacturer_product_root: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAU732#001",
    })
    manufacturer_product_family: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAU731#001",
    })
    manufacturer_product_type: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO057#002",
    })
    order_code_of_manufacturer: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO227#002",
    })
    product_article_number_of_manufacturer: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAO676#003"
    })
    serial_number: Optional[str] = field(metadata={
        "semantic_id": "0173-1#02-AAM556#002",
    })
    year_of_construction: str = field(metadata={
        "semantic_id": "0173-1#02-AAP906#001",
    })
    date_of_manufacture: Optional[datetime.date] = field(metadata={
        "semantic_id": "0173-1#02-AAR972#002",
    })
    hardware_version: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAN270#002",
    })
    firmware_version: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAM985#002",
    })
    software_version: Optional[LangString] = field(metadata={
        "semantic_id": "0173-1#02-AAM737#002",
    })
    country_of_origin: Optional[str] = field(metadata={
        "semantic_id": "0173-1#02-AAO259#004"
    })
    #TODO: company_logo, markings, asset_specific_properties

    def check_family_or_type_present(self):
        if self.manufacturer_product_family is None and self.manufacturer_product_type is None:
            raise CheckConstraintException("Either ManufacturerProductFamily or ManufacturerProductType is required")


def parse_submodel_templates(root_result: AasTestResult, env: Environment):
    for submodel in env.submodels or []:
        if not submodel.semantic_id or not submodel.semantic_id.keys:
            continue
        sid = submodel.semantic_id.keys[0].value.raw_value
        sub_result = AasTestResult(f"Check submodel '{submodel.id}'")
        try:
            template = templates[sid]
        except KeyError:
            sub_result.append(AasTestResult(f"Unknown semantic id '{sid}'", level=Level.WARNING))
            continue
        sub_result.append(AasTestResult(f"Template: {template.__name__} ({sid})"))
        parsed_submodel = parse_submodel(sub_result, template, submodel)
        if sub_result.ok():
            check_constraints(parsed_submodel, sub_result, AdapterPath())
        root_result.append(sub_result)


def supported_templates() -> List[str]:
    return list(templates.keys())



from typing import Dict, List, Optional
from xml.etree.ElementTree import Element


class AdapterPath:
    def __init__(self):
        self.elements = []

    def __add__(self, other):
        result = AdapterPath()
        result.elements = self.elements + [other]
        return result

    def __str__(self):
        return "/" + "/".join([str(i) for i in self.elements])


class AdapterException(Exception):
    pass


class Adapter:

    path: AdapterPath

    def as_object(self) -> Dict[str, "Adapter"]:
        raise NotImplementedError()

    def as_list(self) -> List["Adapter"]:
        raise NotImplementedError()

    def as_string(self) -> str:
        raise NotImplementedError()

    def as_bool(self) -> bool:
        raise NotImplementedError()

    def get_model_type(self) -> str:
        raise NotImplementedError()


class JsonAdapter(Adapter):

    def __init__(self, value: any, path: AdapterPath):
        self.value = value
        self.path = path

    def as_object(self) -> Dict[str, Adapter]:
        if not isinstance(self.value, dict):
            raise AdapterException(f"Cannot convert {self.value} to object")
        return {k: JsonAdapter(v, self.path + k) for k, v in self.value.items() if k != 'modelType'}

    def as_list(self) -> List[Adapter]:
        if not isinstance(self.value, list):
            raise AdapterException(f"Cannot convert {self.value} to list")
        if len(self.value) == 0:
            raise AdapterException(f"Empty array not allowed")
        return [JsonAdapter(val, self.path + idx) for idx, val in enumerate(self.value)]

    def as_string(self):
        if not isinstance(self.value, str):
            raise AdapterException(f"Cannot convert {self.value} to string")
        return self.value

    def as_bool(self):
        if not isinstance(self.value, bool):
            raise AdapterException(f"Cannot convert {self.value} to bool")
        return self.value

    def get_model_type(self) -> str:
        if not isinstance(self.value, dict):
            raise AdapterException(f"Expected an object, got {self.value}")
        try:
            return self.value['modelType']
        except KeyError:
            raise AdapterException(f"Missing 'modelType'")

    def __str__(self):
        return str(self.value)


_expected_namespace = '{https://admin-shell.io/aas/3/0}'


def _get_single_child(el: Element) -> Element:
    if len(el) != 1:
        raise AdapterException("DataSpecificationContent must have exactly one child")
    return el[0]


def _assert_no_children(el: Element):
    if next(iter(el), None):
        raise AdapterException("No child elements allowed")


def _assert_no_text(el: Element):
    if el.text is None:
        return
    if el.text.strip():
        raise AdapterException("No inline text allowed")


def _get_model_type(el: Element, expected_namespace: str):
    model_type = el.tag[len(expected_namespace):]
    model_type = model_type[0].upper() + model_type[1:]
    return model_type


class XmlAdapter:

    def __init__(self, value: Element, path: AdapterPath):
        self.value = value
        self.path = path

    def as_object(self) -> Dict[str, "Adapter"]:
        if not self.value.tag.startswith(_expected_namespace):
            raise AdapterException(f"invalid namespace, got '{self.value.tag}'")
        _assert_no_text(self.value)

        # Special handling for data specification content
        if self.value.tag.endswith('dataSpecificationContent'):
            data = _get_single_child(self.value)
        else:
            data = self.value

        result = {}
        for child in data:
            if not child.tag.startswith(_expected_namespace):
                raise AdapterException(f"invalid namespace, got {child.tag}")
            tag = child.tag[len(_expected_namespace):]
            if self.get_model_type() == 'OperationVariable' and tag == 'value':
                result[tag] = XmlAdapter(_get_single_child(child), self.path + tag)
            else:
                result[tag] = XmlAdapter(child, self.path + tag)
        return result

    def as_list(self) -> List["Adapter"]:
        result = []
        for idx, child in enumerate(self.value):
            result.append(XmlAdapter(child, self.path + idx))
        if not result:
            raise AdapterException("Empty list not allowed")
        return result

    def as_string(self) -> str:
        _assert_no_children(self.value)
        return self.value.text or ""

    def as_bool(self) -> bool:
        _assert_no_children(self.value)
        return self.value.text == 'true'

    def get_model_type(self) -> str:
        # Special handling for data specification content
        if self.value.tag.endswith('dataSpecificationContent'):
            data = _get_single_child(self.value)
        else:
            data = self.value
        return _get_model_type(data, _expected_namespace)



from unittest import TestCase

from aas_test_engines import api
from aas_test_engines.test_cases.v3_0.api import check_in_sync


class SupportedVersionsTest(TestCase):

    def test_list(self):
        s = api.supported_versions()
        for i in s:
            print(i)
        self.assertIn(api.latest_version(), s)

    def test_in_sync(self):
        check_in_sync()



from .parse import StringFormattedValue, abstract, CheckConstraintException, requires_model_type

from dataclasses import dataclass, field
from typing import List, Optional, Set, Iterator, Union
from enum import Enum
from .data_types import _is_bounded_integer, is_bcp_lang_string, DataTypeDefXsd, validators, is_xs_date_time_utc, is_bcp_47_for_english, is_any_uri, is_xs_date_time

# TODO: AASd-021
# TODO: AASd-022
# TODO: AASd-077


class IdShortPath():

    def __init__(self, root: "Submodel"):
        self.root = root
        self.id_shorts: List[Union["NameTypeString", int, None]] = []

    def __add__(self, token: Union["NameTypeString", int, None]):
        result = IdShortPath(self.root)
        result.id_shorts = self.id_shorts + [token]
        return result

    def __str__(self):
        def token_to_string(t):
            if t is None:
                return '?'
            if isinstance(t, int):
                return str(t)
            if isinstance(t, NameTypeString):
                return t.raw_value
            raise Exception(f"Internal error: invalid token '{t}'")
        path = '.'.join(token_to_string(i) for i in self.id_shorts)
        id_short = self.root.id_short.raw_value if self.root.id_short else ""
        return f"{path} in Submodel {id_short}[{self.root.id}]"


def validate(value: str, value_type: DataTypeDefXsd, path: Optional[IdShortPath] = None):
    validator = validators[value_type]
    if not validator(value):
        msg = f"Value '{value}' is not a '{value_type.value}'"
        if path:
            msg += f" @ {path}"
        raise CheckConstraintException(msg)


# 5.3.11.2 Primitive Data Types


@dataclass
class LangStringSet:
    language: str
    text: str

    def __init_subclass__(cls, max_len_text):
        cls._max_len_text = max_len_text
        return super().__init_subclass__()

    def check_language(self):
        if not is_bcp_lang_string(self.language):
            raise CheckConstraintException("Constraint violated: Property 'language' does not contain a language according to BCP46")
        if len(self.language) < 1:
            raise CheckConstraintException("Constraint violated: Property 'language' must not be empty")

    def check_text(self):
        if len(self.text) < 1:
            raise CheckConstraintException("Constraint violated: Property 'text' must not be empty")
        if len(self.text) > self._max_len_text:
            raise CheckConstraintException(f"Constraint violated: Property 'text' is too long ({len(self.text)} > {self._max_len_text}")


class MultiLanguageNameType(LangStringSet, max_len_text=128):
    pass


class MultiLanguageTextType(LangStringSet, max_len_text=1023):
    pass


class BlobString(StringFormattedValue):
    min_length = 1


class ContentType(StringFormattedValue):
    min_length = 1
    max_length = 100
    pattern = r"^([!#$%&'*+\-.^_`|~0-9a-zA-Z])+/([!#$%&'*+\-.^_`|~0-9a-zA-Z])+([        \t]*;[ \t]*([!#$%&'*+\-.^_`|~0-9a-zA-Z])+=(([!#$%&'*+\-.^_`|~0-9a-zA-Z])+|\"(([\t        !#-\[\]-~]|[\x80-ÿ])|\\([\t !-~]|[\x80-ÿ]))*\"))*$"


class IdentifierString(StringFormattedValue):
    min_length = 1
    max_length = 2000


class LabelString(StringFormattedValue):
    min_length = 1
    max_length = 64


class MessageTopicString(StringFormattedValue):
    min_length = 1
    max_length = 255


class NameTypeString(StringFormattedValue):
    min_length = 1
    max_length = 128
    pattern = r"[a-zA-Z][a-zA-Z0-9_]*"


class PathString(StringFormattedValue):
    min_length = 1
    max_length = 2000

    def __init__(self, raw_value):
        super().__init__(raw_value)
        if not is_any_uri(raw_value):
            raise ValueError("Not a valid path")


class RevisionString(StringFormattedValue):
    min_length = 1
    max_length = 4
    pattern = r"(0|[1-9][0-9]*)"


QualifierType = NameTypeString


class VersionString(StringFormattedValue):
    min_length = 1
    max_length = 4
    pattern = r"(0|[1-9][0-9]*)"


class ValueDataType(StringFormattedValue):
    pass


class DateTime(StringFormattedValue):
    min_length = 1

    def __init__(self, raw_value):
        super().__init__(raw_value)
        if not is_xs_date_time(self.raw_value):
            raise ValueError("Not an xs:dateTime")


class DateTimeUtc(StringFormattedValue):
    min_length = 1

    def __init__(self, raw_value):
        super().__init__(raw_value)
        if not is_xs_date_time_utc(self.raw_value):
            raise ValueError("Not an xs:dateTimeUtc")


class Duration(StringFormattedValue):
    min_length = 1
    pattern = r"-?P((([0-9]+Y([0-9]+M)?([0-9]+D)?|([0-9]+M)([0-9]+D)?|([0-9]+D))(T(([0-9]+H)([0-9]+M)?([0-9]+(\.[0-9]+)?S)?|([0-9]+M)([0-9]+(\.[0-9]+)?S)?|([0-9]+(\.[0-9]+)?S)))?)|(T(([0-9]+H)([0-9]+M)?([0-9]+(\.[0-9]+)?S)?|([0-9]+M)([0-9]+(\.[0-9]+)?S)?|([0-9]+(\.[0-9]+)?S))))"


class DataType(StringFormattedValue):
    min_length = 1


class ValueType(StringFormattedValue):
    min_length = 1


# 5.3.12.2 Constraints for Referables and Identifiables


def ensure_have_id_shorts(elements: Optional[List["Referable"]], id_short_path: Optional[IdShortPath] = None):
    """
    Constraint AASd-117: idShort of non-identifiable Referables not being a direct child of a
    SubmodelElementList shall be specified.
    """
    if elements is None:
        return
    for idx, val in enumerate(elements):
        if val.id_short is None:
            msg = f"Constraint AASd-117 violated: element {idx} has no idShort"
            if id_short_path:
                msg += f" @ {id_short_path}"
            raise CheckConstraintException(msg)


# 5.3.10.2 Reference

class KeyType(Enum):
    AnnotatedRelationshipElement = 'AnnotatedRelationshipElement'
    AssetAdministrationShell = 'AssetAdministrationShell'
    BasicEventElement = 'BasicEventElement'
    Blob = 'Blob'
    Capability = 'Capability'
    ConceptDescription = 'ConceptDescription'
    DataElement = 'DataElement'
    Entity = 'Entity'
    EventElement = 'EventElement'
    File = 'File'
    FragmentReference = 'FragmentReference'
    GlobalReference = 'GlobalReference'
    Identifiable = 'Identifiable'
    MultiLanguageProperty = 'MultiLanguageProperty'
    Operation = 'Operation'
    Property = 'Property'
    Range = 'Range'
    Referable = 'Referable'
    ReferenceElement = 'ReferenceElement'
    RelationshipElement = 'RelationshipElement'
    Submodel = 'Submodel'
    SubmodelElement = 'SubmodelElement'
    SubmodelElementCollection = 'SubmodelElementCollection'
    SubmodelElementList = 'SubmodelElementList'

    def matches(self, obj) -> bool:
        return obj.__class__.__name__ == self.value


AasSubmodelElements = {
    KeyType.AnnotatedRelationshipElement,
    KeyType.BasicEventElement,
    KeyType.Blob,
    KeyType.Capability,
    KeyType.DataElement,
    KeyType.Entity,
    KeyType.EventElement,
    KeyType.File,
    KeyType.MultiLanguageProperty,
    KeyType.Operation,
    KeyType.Property,
    KeyType.Range,
    KeyType.ReferenceElement,
    KeyType.RelationshipElement,
    KeyType.SubmodelElement,
    KeyType.SubmodelElementCollection,
    KeyType.SubmodelElementList,
}

GenericGloballyIdentifiables = {
    KeyType.GlobalReference,
}

AasIdentifiables = {
    KeyType.AssetAdministrationShell,
    KeyType.ConceptDescription,
    KeyType.Identifiable,
    KeyType.Submodel,
}

GenericFragmentKeys = {
    KeyType.FragmentReference,
}

GloballyIdentifiables = AasIdentifiables | GenericGloballyIdentifiables

FragmentKeys = GenericFragmentKeys | AasSubmodelElements


@dataclass
class Key:
    type: KeyType
    value: IdentifierString

    def __eq__(self, other: "Key") -> bool:
        return self.type == other.type and self.value == other.value

    def __str__(self):
        return self.value.raw_value


class ReferenceType(Enum):
    ExternalReference = 'ExternalReference'
    ModelReference = 'ModelReference'


@dataclass
class Reference:
    type: ReferenceType
    referred_semantic_id: Optional["Reference"]
    keys: List[Key]

    def check_aasd_121(self):
        """
        Constraint AASd-121: For References, the value of Key/type of the first key of Reference/keys shall be one
        of GloballyIdentifiables.
        """
        if not self.keys:
            return
        if self.keys[0].type not in GloballyIdentifiables:
            raise CheckConstraintException("Constraint AASd-121 violated: first key must be one of GloballyIdentifiables")

    def check_aasd_122(self):
        """
        Constraint AASd-122: For external references, i.e. References with Reference/type = ExternalReference, the
        value of Key/type of the first key of Reference/keys shall be one of GenericGloballyIdentifiables.
        """
        if self.type != ReferenceType.ExternalReference or self.keys is None:
            return
        if self.keys[0].type not in GenericGloballyIdentifiables:
            raise CheckConstraintException("Constraint AASd-122 violated: first key must be one of GenericGloballyIdentifiables")

    def check_aasd_123(self):
        """
        Constraint AASd-123: For model references, i.e. References with Reference/type = ModellReference, the
        value of Key/type of the first key of Reference/keys shall be one of AasIdentifiables.
        """
        if self.type != ReferenceType.ModelReference or self.keys is None:
            return
        if self.keys[0].type not in AasIdentifiables:
            raise CheckConstraintException("Constraint AASd-123 violated: first key must be one of AasIdentifiables")

    def check_aasd_124(self):
        """
        Constraint AASd-124: For external references, i.e. References with Reference/type = ExternalReference, the
        last key of Reference/keys shall be either one of GenericGloballyIdentifiables or one of
        GenericFragmentKeys.
        """
        if self.type != ReferenceType.ExternalReference or self.keys is None:
            return
        if self.keys[-1].type not in GenericGloballyIdentifiables and self.keys[-1].type not in GenericFragmentKeys:
            raise CheckConstraintException("Constraint AASd-124 violated: last key must one of GenericGloballyIdentifiables or GenericFragmentKeys")

    def check_aasd_125(self):
        """
        Constraint AASd-125: For model references, i.e. References with Reference/type = ModelReference with
        more than one key in Reference/keys, the value of Key/type of each of the keys following the first key of
        Reference/keys shall be one of FragmentKeys.
        """
        if self.type != ReferenceType.ModelReference or self.keys is None:
            return
        for idx, key in enumerate(self.keys[1:]):
            if key.type not in FragmentKeys:
                raise CheckConstraintException(f"Constraint AASd-125 violated: key {idx} must be a FragmentKey")

    def check_aasd_126(self):
        """
        Constraint AASd-126: For model references, i.e. References with Reference/type = ModelReference with
        more than one key in Reference/keys, the value of Key/type of the last Key in the reference key chain may
        be one of GenericFragmentKeys or no key at all shall have a value out of GenericFragmentKeys.
        """
        if self.type != ReferenceType.ModelReference or self.keys is None or len(self.keys) <= 1:
            return
        for idx, key in enumerate(self.keys[:-1]):
            if key.type in GenericFragmentKeys:
                raise CheckConstraintException(f"Constraint AASd-126 violated: key {idx} must not be a GenericFragmentKey")

    def check_aasd_127(self):
        """
        Constraint AASd-127: For model references, i.e. References with Reference/type = ModelReference with
        more than one key in Reference/keys, a key with Key/type FragmentReference shall be preceded by a key
        with Key/type File or Blob. All other Asset Administration Shell fragments, i.e. Key/type values out of
        AasSubmodelElements, do not support fragments.
        """
        if self.type != ReferenceType.ModelReference or self.keys is None or len(self.keys) <= 1:
            return
        for idx, key in enumerate(self.keys):
            if key.type == KeyType.FragmentReference and idx >= 1:
                if self.keys[idx-1].type not in {KeyType.File, KeyType.Blob}:
                    raise CheckConstraintException(f"Constraint AASd-127 violated: key {idx} must be preceded by File or Blob")

    def check_aasd_128(self):
        """
        Constraint AASd-128: For model references, i.e. References with Reference/type = ModelReference, the
        Key/value of a Key preceded by a Key with Key/type=SubmodelElementList is an integer number denoting
        the position in the array of the submodel element list. 
        """
        if self.type != ReferenceType.ModelReference or self.keys is None:
            return
        for idx, key in enumerate(self.keys):
            if key.type == KeyType.SubmodelElementList and idx < len(self.keys)-1:
                if not _is_bounded_integer(self.keys[idx+1].value.raw_value, 0, float('inf')):
                    raise CheckConstraintException(f"Constraint AASd-128 violated: key {idx} must be followed by an integer")

    def __eq__(self, other: "Reference") -> bool:
        if self.type != other.type:
            return False
        if len(self.keys) != len(other.keys):
            return False
        for x, y in zip(self.keys, other.keys):
            if x != y:
                return False
        return True

    def __str__(self) -> str:
        result = self.type.value
        if self.keys:
            result += "/".join([str(key) for key in self.keys])
        return result

# 5.3.2.3 Has Data Specification


@dataclass
@abstract
class DataSpecificationContent:
    pass


@dataclass
class EmbeddedDataSpecification:
    data_specification: Reference
    data_specification_content: DataSpecificationContent


@dataclass
class HasDataSpecification:
    embedded_data_specifications: Optional[List[EmbeddedDataSpecification]]


@dataclass
class ValueReferencePair:
    value: IdentifierString
    value_id: Reference


@dataclass
class ValueList:
    value_reference_pairs: List[ValueReferencePair]


@dataclass
class LevelType:
    min: bool
    nom: bool
    typ: bool
    max: bool


class DataTypeIec61360(Enum):
    BLOB = "BLOB"
    BOOLEAN = "BOOLEAN"
    DATE = "DATE"
    FILE = "FILE"
    HTML = "HTML"
    INTEGER_COUNT = "INTEGER_COUNT"
    INTEGER_CURRENCY = "INTEGER_CURRENCY"
    INTEGER_MEASURE = "INTEGER_MEASURE"
    IRDI = "IRDI"
    IRI = "IRI"
    RATIONAL = "RATIONAL"
    RATIONAL_MEASURE = "RATIONAL_MEASURE"
    REAL_COUNT = "REAL_COUNT"
    REAL_CURRENCY = "REAL_CURRENCY"
    REAL_MEASURE = "REAL_MEASURE"
    STRING = "STRING"
    STRING_TRANSLATABLE = "STRING_TRANSLATABLE"
    TIME = "TIME"
    TIMESTAMP = "TIMESTAMP"


class DefinitionTypeIec61360(LangStringSet, max_len_text=1023):
    pass


class PreferredNameTypeIec61360(LangStringSet, max_len_text=255):
    pass


class ShortNameTypeIec61360(LangStringSet, max_len_text=18):
    pass


class ValueTypeIec61360(StringFormattedValue):
    min_length = 1
    max_length = 2000


class NonEmptyString(StringFormattedValue):
    min_length = 1


@dataclass
class DataSpecificationIec61360(DataSpecificationContent):
    preferred_name: List[PreferredNameTypeIec61360]
    short_name: Optional[List[ShortNameTypeIec61360]]
    unit: Optional[NonEmptyString]
    unit_id: Optional[Reference]
    source_of_definition: Optional[NonEmptyString]
    symbol: Optional[NonEmptyString]
    data_type: Optional[DataTypeIec61360]
    definition: Optional[List[DefinitionTypeIec61360]]
    value_format: Optional[NonEmptyString]
    value_list: Optional[ValueList]
    value: Optional[ValueTypeIec61360]
    level_type: Optional[LevelType]

    def check_aasc_3a_002(self):
        """
        Constraint AASc-3a-002: DataSpecificationIec61360/preferredName shall be provided at
        least in English.
        """
        if not any(is_bcp_47_for_english(i.language) for i in self.preferred_name):
            raise CheckConstraintException("Constraint AASc-3a-002 violated: English language is missing")

    def check_aasc_3a_009(self):
        """
        Constraint AASc-3a-009: If DataSpecificationIec61360/dataType is one of
        INTEGER_MEASURE, REAL_MEASURE, RATIONAL_MEASURE,
        INTEGER_CURRENCY, REAL_CURRENCY, then DataSpecificationIec61360/unit or
        DataSpecificationIec61360/unitId shall be defined.
        """
        types = {
            DataTypeIec61360.INTEGER_MEASURE,
            DataTypeIec61360.REAL_MEASURE,
            DataTypeIec61360.RATIONAL_MEASURE,
            DataTypeIec61360.INTEGER_CURRENCY,
            DataTypeIec61360.REAL_CURRENCY,
        }
        if self.data_type in types and (self.unit is None and self.unit_id is None):
            raise CheckConstraintException("Constraint AASc-3a-009 violated: neither unit nor unit_id is set")

    def check_aasc_3a_010(self):
        """
        Constraint AASc-3a-010: If DataSpecificationIec61360/value is not empty,
        DataSpecificationIec61360/valueList shall be empty, and vice versa
        """
        if self.value is not None and self.value_list is not None:
            raise CheckConstraintException("Constraint AASc-3a-010 violated: value and value_list cannot be set simultaneously")


# 5.3.2.6 Has Semantics


@dataclass
class HasSemantics:
    # TODO: Note: it is recommended to use an external reference
    semantic_id: Optional[Reference]
    # TODO: Note: it is recommended to use an external reference
    supplemental_semantic_ids: Optional[List[Reference]]

    def check_aasd_118(self):
        """
        Constraint AASd-118: If a supplemental semantic ID
        (HasSemantics/supplementalSemanticId) is defined, there shall also be a main
        semantic ID (HasSemantics/semanticId)
        """
        if self.supplemental_semantic_ids is not None and self.semantic_id is None:
            raise CheckConstraintException("Constraint AASd-118 violated: supplementalSemanticId is given but semanticId is missing")

# 5.3.2.4 Extensions


@dataclass
class Extension(HasSemantics):
    name: NameTypeString
    # TODO: default: xs:string
    value_type: Optional[DataTypeDefXsd]
    value: Optional[ValueDataType]
    # TODO: must be a model reference
    refers_to: Optional[List[Reference]]

    def check_value_type(self):
        if self.value and self.value_type:
            validate(self.value.raw_value, self.value_type)


@dataclass
class HasExtensions:
    extensions: Optional[List[Extension]]

# 5.3.2.10 Referable


@dataclass
class Referable(HasExtensions):
    # TODO: category is deprecated
    category: Optional[NameTypeString]
    id_short: Optional[NameTypeString]
    display_name: Optional[List[MultiLanguageNameType]]
    description: Optional[List[MultiLanguageTextType]]
    # TODO: Constraint AASd-002: idShort of Referables shall only feature letters, digits, underscore ("_");
    # starting mandatory with a letter, i.e. [a-zA-Z][a-zA-Z0-9_]*.


# 5.3.2.5 Has Kind


class ModellingKind(Enum):
    Instance = "Instance"
    Template = "Template"


@dataclass
class HasKind:
    kind: Optional[ModellingKind]

# 5.3.2.2 Administrative Information


@dataclass
class AdministrativeInformation(HasDataSpecification):
    version: Optional[VersionString]
    revision: Optional[RevisionString]
    creator: Optional[Reference]
    template_id: Optional[IdentifierString]

    def check_constraint_aasd_005(self):
        """
        Constraint AASd-005: If AdministrativeInformation/version is not specified,
        AdministrativeInformation/revision shall also be unspecified. This means that a revision
        requires a version. If there is no version, there is no revision. Revision is optional.
        """
        if self.revision is not None and self.version is None:
            raise CheckConstraintException("Constraint AASd-005 violated: version is given but no revision")

# 5.3.2.7 Identifiable


@dataclass
@requires_model_type
class Identifiable(Referable):
    administration: Optional[AdministrativeInformation]
    id: IdentifierString

# 5.3.2.8 Qualifiable


class QualifierKind(Enum):
    ConceptQualifier = 'ConceptQualifier'
    TemplateQualifier = 'TemplateQualifier'
    ValueQualifier = 'ValueQualifier'


@dataclass
class Qualifier(HasSemantics):
    # TODO: Default: ConceptQualifier
    kind: Optional[QualifierKind]
    type: QualifierType
    value_type: DataTypeDefXsd
    value: Optional[ValueDataType]
    # TODO: Note: it is recommended to use an external reference.
    value_id: Optional[Reference]

    def check_aasd_006(self):
        """
        Constraint AASd-006: If both, the value and the valueId of a Qualifier are present, the value
        needs to be identical to the value of the referenced coded value in Qualifier/valueId.
        => not checkable
        """
        pass

    def check_aasd_020(self):
        """
        Constraint AASd-020: The value of Qualifier/value shall be consistent with the data type as
        defined in Qualifier/valueType.
        """
        # TODO
        pass

    def check_value_type(self):
        if self.value:
            validate(self.value.raw_value, self.value_type)


@dataclass
class Qualifiable:
    qualifiers: Optional[List[Qualifier]]

# 5.3.4 Asset Information


@dataclass
class SpecificAssetId(HasSemantics):
    name: LabelString
    value: IdentifierString
    external_subject_id: Optional[Reference]

    def check_aasd_133(self):
        """
        Constraint AASd-133: SpecificAssetId/externalSubjectId shall be a global reference, i.e.
        Reference/type = ExternalReference.
        """
        if self.external_subject_id and self.external_subject_id.type != ReferenceType.ExternalReference:
            raise CheckConstraintException("Constraint AASd-133 violated: type must be ExternalReference")


@dataclass
class Resource:
    path: PathString
    content_type: Optional[ContentType]


class AssetKind(Enum):
    INSTANCE = 'Instance'
    NOT_APPLICABLE = 'NotApplicable'
    TYPE = 'Type'


@dataclass
class AssetInformation:
    asset_kind: AssetKind
    global_asset_id: Optional[IdentifierString]
    specific_asset_ids: Optional[List[SpecificAssetId]]
    asset_type: Optional[IdentifierString]
    default_thumbnail: Optional[Resource]

    def check_aasd_131(self):
        """
        Constraint AASd-131: The globalAssetId or at least one specificAssetId shall be defined for
        AssetInformation.
        """
        if not self.global_asset_id and not self.specific_asset_ids:
            raise CheckConstraintException("Constraint AASd-131 violated: neither globalAssetId nor specificAssetIds given")

    def check_aasd_116(self):
        """
        Constraint AASd-116: : "globalAssetId" (case-insensitive) is a reserved key for SpecificAssetId/name with the
        semantics as defined in https://admin-shell.io/aas/3/0/AssetInformation/globalAssetId.
        """
        if self.specific_asset_ids is None:
            return
        for idx, specific_asset_id in enumerate(self.specific_asset_ids):
            if specific_asset_id.name.raw_value == 'globalAssetId':
                raise CheckConstraintException(f"Constraint AASd-116 violated: specific asset id {idx} must not have name 'globalAssetId'")

# 5.3.6 Submodel Element


@dataclass
@abstract
class SubmodelElement(Referable, HasSemantics, Qualifiable, HasDataSpecification):
    id_short_path: IdShortPath = field(metadata={'exclude_as': None})

    def elements(self) -> Iterator["SubmodelElement"]:
        yield self

    def _set_path(self, path: IdShortPath):
        self.id_short_path = path

    def check_aasc_3a_050(self):
        """
        Constraint AASc-3a-050: If the DataSpecificationContent DataSpecificationIec61360 is used for an element,
        the value of HasDataSpecification/dataSpecification shall contain the external reference to the IRI of the
        corresponding data specification template https://adminshell.io/DataSpecificationTemplates/DataSpecificationIec61360/3.
        """
        # TODO


# 5.3.7.6 Data Element


@dataclass
@abstract
class DataElement(SubmodelElement):

    def check_aasd_090(self):
        """
        Constraint AASd-090: for data elements, category (inherited by Referable) shall be one of
        the following values: CONSTANT, PARAMETER or VARIABLE. Default: VARIABLE
        """
        allowed_values = ["CONSTANT", "PARAMETER", "VARIABLE"]
        if self.category is not None and self.category.raw_value not in allowed_values:
            raise CheckConstraintException(f"Constraint AASd-090 violated: category {self.category.raw_value} is not one of {allowed_values} @ {self.id_short_path}")


# 5.3.7.15 Relationship Element

@dataclass
class RelationshipElement(SubmodelElement):
    first: Reference
    second: Reference

# 5.3.7.2 Annotated Relationship Element


@dataclass
class AnnotatedRelationshipElement(RelationshipElement):
    annotations: Optional[List[DataElement]]

    def check_aasd_117(self):
        ensure_have_id_shorts(self.annotations, self.id_short_path)


# 5.3.7.8 Event Element

@dataclass
class EventElement(SubmodelElement):
    pass

# 5.3.7.3 Basic Event Element


class StateOfEvent(Enum):
    OFF = 'off'
    ON = 'on'


class Direction(Enum):
    INPUT = 'input'
    OUTPUT = 'output'


@dataclass
class BasicEventElement(EventElement):
    observed: Reference
    direction: Direction
    state: StateOfEvent
    message_topic: Optional[MessageTopicString]
    message_broker: Optional[Reference]
    last_update: Optional[DateTime]
    min_interval: Optional[Duration]
    max_interval: Optional[Duration]

    def check_observed(self):
        if self.observed.type != ReferenceType.ModelReference:
            raise CheckConstraintException(f"Constraint violated: Property 'observed' must be a model reference @ {self.id_short_path}")

    def check_message_broker(self):
        if self.message_broker and self.message_broker.type != ReferenceType.ModelReference:
            raise CheckConstraintException(f"Constraint violated: Property 'observed' must be a model reference @ {self.id_short_path}")

# 5.3.7.4 Blob


@dataclass
class Blob(DataElement):
    value: Optional[BlobString]
    content_type: ContentType

# 5.3.7.5 Capability


@dataclass
class Capability(SubmodelElement):
    pass


# 5.3.7.7 Entity

class EntityType(Enum):
    CoManagedEntity = 'CoManagedEntity'
    SelfManagedEntity = 'SelfManagedEntity'


@dataclass
class Entity(SubmodelElement):
    statements: Optional[List[SubmodelElement]]
    entity_type: EntityType
    global_asset_id: Optional[IdentifierString]
    specific_asset_ids: Optional[List[SpecificAssetId]]

    def check_aasd_014(self):
        """
        Constraint AASd-014: Either the attribute globalAssetId or specificAssetId of an Entity must
        be set if Entity/entityType is set to "SelfManagedEntity". Otherwise, they do not exist.
        """
        if self.entity_type == EntityType.SelfManagedEntity:
            if self.global_asset_id is None and self.specific_asset_ids is None:
                raise CheckConstraintException(f"Constraint AASd-014 violated: entity is self-manged but neither globalAssetId nor specificAssetId are set @ {self.id_short_path}")
        else:
            if self.global_asset_id or self.specific_asset_ids:
                raise CheckConstraintException(f"Constraint AASd-014 violated: entity is co-manged but either globalAssetId or specificAssetId are set @ {self.id_short_path}")

    def check_aad_117(self):
        ensure_have_id_shorts(self.statements, self.id_short_path)

# 5.3.7.9 File


@dataclass
class File(DataElement):
    value: Optional[PathString]
    content_type: ContentType


# 5.3.7.10 Multi Language Property


@dataclass
class MultiLanguageProperty(DataElement):
    value: Optional[List[MultiLanguageTextType]]
    # TODO: Note: it is recommended to use an external reference.
    value_id: Optional[Reference]

    def check_constraint_aasd_012(self):
        """
        Constraint AASd-012: if both the MultiLanguageProperty/value and the
        MultiLanguageProperty/valueId are present, the meaning must be the same for each string
        in a specific language, as specified in MultiLanguageProperty/valueId.
        => not checked
        """
        pass

# 5.3.7.11 Operation


@dataclass
class OperationVariable:
    value: SubmodelElement

    def check_aasd_117(self):
        """
        Constraint AASd-117: idShort of non-identifiable Referables not being a direct child of a
        SubmodelElementList shall be specified.
        """
        if self.value.id_short is None:
            raise CheckConstraintException(f"Constraint AASd-117 is violated: idShort missing @ {self.id_short_path}")


@dataclass
class Operation(SubmodelElement):
    input_variables: Optional[List[OperationVariable]]
    output_variables: Optional[List[OperationVariable]]
    inoutput_variables: Optional[List[OperationVariable]]

    def _check_list(self, l: Optional[List[OperationVariable]], all_id_shorts: Set[str]):
        if l is None:
            return
        for i in l:
            if i.value.id_short is None:
                continue
            if i.value.id_short.raw_value in all_id_shorts:
                raise CheckConstraintException(f"Constraint AASd-134 violated: duplicate id short @ {self.id_short_path}")
            all_id_shorts.add(i.value.id_short.raw_value)

    def check_aasd_134(self):
        """
        Constraint AASd-134: For an Operation, the idShort of all inputVariable/value,
        outputVariable/value, and inoutputVariable/value shall be unique.
        """
        all_id_shorts = set()
        self._check_list(self.input_variables, all_id_shorts)
        self._check_list(self.output_variables, all_id_shorts)
        self._check_list(self.inoutput_variables, all_id_shorts)

# 5.3.7.12 Property


@dataclass
class Property(DataElement):
    value_type: DataTypeDefXsd
    value: Optional[ValueDataType]
    # TODO: Note: it is recommended to use an external reference.
    value_id: Optional[Reference]

    def check_aasd_007(self):
        """
        Constraint AASd-007: If both the Property/value and the Property/valueId are present, the
        value of Property/value needs to be identical to the value of the referenced coded value in
        Property/valueId.
        -> cannot check
        """
        pass

    def check_value_type(self):
        if self.value:
            validate(self.value.raw_value, self.value_type, self.id_short_path)

# 5.3.7.13 Range


@dataclass
class Range(DataElement):
    value_type: DataTypeDefXsd
    min: Optional[ValueDataType]
    max: Optional[ValueDataType]

    def check_value_type_min(self):
        if self.min:
            validate(self.min.raw_value, self.value_type, self.id_short_path)

    def check_value_type_max(self):
        if self.max:
            validate(self.max.raw_value, self.value_type, self.id_short_path)

# 5.3.7.14 Reference Element


@dataclass
class ReferenceElement(DataElement):
    value: Optional[Reference]


# 5.3.7.16 Submodel Element Collection

@dataclass
class SubmodelElementCollection(SubmodelElement):
    value: Optional[List[SubmodelElement]]

    def elements(self) -> Iterator["SubmodelElement"]:
        yield self
        if self.value:
            for el in self.value:
                yield from el.elements()

    def _set_path(self, path):
        self.id_short_path = path
        if self.value:
            for el in self.value:
                el._set_path(path + el.id_short)

    def check_aasd_117(self):
        ensure_have_id_shorts(self.value, self.id_short_path)

# 5.3.7.17 Submodel Element List


@dataclass
class SubmodelElementList(SubmodelElement):
    # TODO: default: true
    order_relevant: Optional[bool]
    # TODO: Note: it is recommended to use an external reference.
    semantic_id_list_element: Optional[Reference]
    type_value_list_element: KeyType
    value_type_list_element: Optional[DataTypeDefXsd]
    value: Optional[List[SubmodelElement]]

    def _set_path(self, path):
        self.id_short_path = path
        if self.value:
            for idx, el in enumerate(self.value):
                el._set_path(path + idx)

    def elements(self) -> Iterator["SubmodelElement"]:
        yield self
        if self.value:
            for el in self.value:
                yield from el.elements()

    def check_type_value_list_element(self):
        if self.type_value_list_element not in AasSubmodelElements:
            raise CheckConstraintException(f"Constraint violated: type_value_list_element must be a AasSubmodelElement @ {self.id_short_path}")

    def check_aasd_107(self):
        """
        Constraint AASd-107: If a first level child element in a SubmodelElementList has a
        semanticId, it shall be identical to SubmodelElementList/semanticIdListElement.
        """
        if not self.value or not self.semantic_id_list_element:
            return
        for idx, el in enumerate(self.value):
            if el.semantic_id is None:
                continue
            if el.semantic_id != self.semantic_id_list_element:
                raise CheckConstraintException(f"Constraint AASd-107 violated: Element {idx} has invalid semantic id {el.semantic_id}, should be {self.semantic_id_list_element} @ {self.id_short_path}")

    def check_aasd_114(self):
        """
        Constraint AASd-114: If two first level child elements in a SubmodelElementList
        have a semanticId, they shall be identical.
        """
        if not self.value:
            return
        semantic_id = None
        for idx, el in enumerate(self.value):
            if not el.semantic_id:
                continue
            if semantic_id:
                if el.semantic_id != semantic_id:
                    raise CheckConstraintException(f"Constraint AASd-114 violated: Element {idx} must have semanticId {semantic_id} @ {self.id_short_path}")
            else:
                semantic_id = el.semantic_id

    def check_aasd_115(self):
        """
        Constraint AASd-115: If a first level child element in a SubmodelElementList does
        not specify a semanticId, the value is assumed to be identical to
        SubmodelElementList/semanticIdListElement.
        -> not checked
        """
        pass

    def check_aasd_108(self):
        """
        Constraint AASd-108: All first level child elements in a SubmodelElementList shall
        have the same submodel element type as specified in
        SubmodelElementList/typeValueListElement.
        """
        if not self.value:
            return
        if not self.type_value_list_element:
            return
        for idx, el in enumerate(self.value):
            if not self.type_value_list_element.matches(el):
                raise CheckConstraintException(f"Constraint AASd-108 violated: Element {idx} must be {self.type_value_list_element.value} @ {self.id_short_path}")

    def check_aasd_109(self):
        """
        Constraint AASd-109: If SubmodelElementList/typeValueListElement is equal to
        Property or Range, SubmodelElementList/valueTypeListElement shall be set and
        all first level child elements in the SubmodelElementList shall have the value type
        as specified in SubmodelElementList/valueTypeListElement.
        """
        if not self.value:
            return
        for idx, el in enumerate(self.value):
            if isinstance(el, (Property, Range)):
                if self.value_type_list_element is None:
                    raise CheckConstraintException(f"Constraint AASd-109 violated: valueTypeListElement must be set since there are Properties/Ranges @ {self.id_short_path}")
                if el.value_type != self.value_type_list_element:
                    raise CheckConstraintException(f"Constraint AASd-109 violated: value type of element {idx} does not match valueTypeListElement @ {self.id_short_path}")

    def check_aasd_120(self):
        """
        Constraint AASd-117: idShort of non-identifiable Referables not being a direct child of a
        SubmodelElementList shall be specified.
        """
        if not self.value:
            return
        for idx, el in enumerate(self.value):
            if el.id_short is not None:
                raise CheckConstraintException(f"Constraint AASd-120 violated: element {idx} must not have an idShort @ {self.id_short_path}")

# 5.3.3 Asset Administration Shell


@dataclass
class AssetAdministrationShell(Identifiable, HasDataSpecification):
    derived_from: Optional[Reference]
    asset_information: AssetInformation
    submodels: Optional[List[Reference]]

    def check_derived_from(self):
        if self.derived_from and self.derived_from.type != ReferenceType.ModelReference:
            raise CheckConstraintException(f"Constraint violated: derivedFrom must be a model reference @ {self.id_short_path}")

    def check_submodels(self):
        if self.submodels:
            for i in self.submodels:
                if i.type != ReferenceType.ModelReference:
                    raise CheckConstraintException(f"Constraint violated: submodels must contain only model references @ {self.id_short_path}")

# 5.3.5 Submodel


@dataclass
class Submodel(Identifiable, HasKind, HasSemantics, Qualifiable, HasDataSpecification):
    submodel_elements: Optional[List[SubmodelElement]]

    def post_parse(self):
        self.update_id_shorts()

    def update_id_shorts(self):
        if self.submodel_elements is None:
            return
        for element in self.submodel_elements:
            element._set_path(IdShortPath(self) + element.id_short)

    def elements(self) -> Iterator[SubmodelElement]:
        if self.submodel_elements is None:
            return
        for element in self.submodel_elements:
            yield from element.elements()

    def check_aasd_117(self):
        ensure_have_id_shorts(self.submodel_elements)

    def check_aasd119(self):
        """
        Constraint AASd-119: If any Qualifier/kind value of a Qualifiable/qualifier is equal to TemplateQualifier and
        the qualified element inherits from "hasKind", the qualified element shall be of kind Template (HasKind/kind =
        "Template").
        """
        if self.qualifiers is None:
            return
        if any(qualifier.kind == QualifierKind.TemplateQualifier for qualifier in self.qualifiers):
            if self.kind != ModellingKind.Template:
                raise CheckConstraintException("Constraint AASd-129 violated: kind must be Template as at least one qualifier is a TemplateQualifier")

    def check_aasd_129(self):
        """
        Constraint AASd-129: If any Qualifier/kind value of a SubmodelElement/qualifier (attribute qualifier inherited
        via Qualifiable) is equal to TemplateQualifier, the submodel element shall be part of a submodel template,
        i.e. a Submodel with Submodel/kind (attribute kind inherited via HasKind) value equal to Template.
        """
        if self.submodel_elements is None:
            return

        for element in self.submodel_elements:
            if element.qualifiers is None:
                continue
            if any(qualifier.kind == QualifierKind.TemplateQualifier for qualifier in element.qualifiers):
                if self.kind != ModellingKind.Template:
                    raise CheckConstraintException("Constraint AASd-129 violated: kind must be Template as at least one qualifier is a TemplateQualifier")


# 5.3.8 Concept Description


@dataclass
class ConceptDescription(Identifiable, HasDataSpecification):
    # TODO: Note: it is recommended to use an external reference, i.e. Reference/type = ExternalReference.
    is_case_of: Optional[List[Reference]]

    def _check_iec_61360(self, categories: Set[str], types, constraint):
        if self.category is None or self.category.raw_value not in categories:
            return
        if self.embedded_data_specifications is None:
            return
        for idx, ds in enumerate(self.embedded_data_specifications):
            if isinstance(ds.data_specification_content, DataSpecificationIec61360):
                if ds.data_specification_content.data_type not in types:
                    raise CheckConstraintException(f"Constraint {constraint} violated: embeddedDataSpecifications[{idx}].dataType is invalid")

    def check_aasc_3a_004(self):
        """
        Constraint AASc-3a-004: For a ConceptDescription with category PROPERTY or VALUE using data
        specification template IEC61360 (https://adminshell.io/DataSpecificationTemplates/DataSpecificationIec61360/3), DataSpecificationIec61360/dataType is
        mandatory and shall be one of DATE, STRING, STRING_TRANSLATABLE, INTEGER_MEASURE,
        INTEGER_COUNT, INTEGER_CURRENCY, REAL_MEASURE, REAL_COUNT, REAL_CURRENCY,
        BOOLEAN, RATIONAL, RATIONAL_MEASURE, TIME, TIMESTAMP.
        """
        self._check_iec_61360({
            'PROPERTY',
            'VALUE'
        }, {
            'DATE',
            'STRING',
            'STRING_TRANSLATABLE',
            'INTEGER_MEASURE',
            'INTEGER_COUNT',
            'INTEGER_CURRENCY',
            'REAL_MEASURE',
            'REAL_COUNT',
            'REAL_CURRENCY',
            'BOOLEAN',
            'RATIONAL',
            'RATIONAL_MEASURE',
            'TIME',
            'TIMESTAMP',
        }, "AASc-3a-004")

    def check_aasc_3a_005(self):
        """
        Constraint AASc-3a-005: For a ConceptDescription with category REFERENCE using data specification
        template IEC61360 (https://admin-shell.io/DataSpecificationTemplates/DataSpecificationIec61360/3),
        DataSpecificationIec61360/dataType shall be one of STRING, IRI, IRDI.
        """
        self._check_iec_61360({
            'REFERENCE',
        }, {
            'STRING',
            'IRI',
            'IRDI',
        }, "AASc-3a-005")

    def check_aasc_3a_006(self):
        """
        Constraint AASc-3a-006: For a ConceptDescription with category DOCUMENT using data specification
        template IEC61360 (https://admin-shell.io/DataSpecificationTemplates/DataSpecificationIec61360/3),
        DataSpecificationIec61360/dataType shall be one of FILE, BLOB, HTML.
        """
        self._check_iec_61360({
            'DOCUMENT',
        }, {
            'FILE',
            'BLOB',
            'HTML',
        }, "AASc-3a-006")

    def check_aasc_3a_007(self):
        """
        Constraint AASc-3a-007: For a ConceptDescription with category QUALIFIER_TYPE using data
        specification template IEC61360 (https://adminshell.io/DataSpecificationTemplates/DataSpecificationIec61360/3), DataSpecificationIec61360/dataType is
        mandatory and shall be defined.
        """
        if self.category is None or self.category.raw_value != 'QUALIFIER_TYPE':
            return
        if self.embedded_data_specifications is None:
            return
        for idx, ds in enumerate(self.embedded_data_specifications):
            if isinstance(ds.data_specification_content, DataSpecificationIec61360):
                if ds.data_specification_content.data_type is None:
                    raise CheckConstraintException(f"Constraint AASc-3a-007 violated: embeddedDataSpecifications[{idx}].dataType is missing")

    def check_aasc_3a_008(self):
        """
        Constraint AASc-3a-008: For a ConceptDescription using data specification template IEC61360
        (https://admin-shell.io/DataSpecificationTemplates/DataSpecificationIec61360/3),
        DataSpecificationIec61360/definition is mandatory and shall be defined at least in English. Exception: the
        concept description describes a value, i.e. DataSpecificationIec61360/value is defined.
        """
        if self.embedded_data_specifications is None:
            return
        for idx, ds in enumerate(self.embedded_data_specifications):
            if isinstance(ds.data_specification_content, DataSpecificationIec61360):
                if ds.data_specification_content.value is not None:
                    continue
                if ds.data_specification_content.definition is None:
                    raise CheckConstraintException(f"Constraint AASc-3a-008 violated: embeddedDataSpecifications[{idx}].definition is missing")
                if not any(is_bcp_47_for_english(i.language) for i in ds.data_specification_content.definition):
                    raise CheckConstraintException(f"Constraint AASc-3a-008 violated: embeddedDataSpecifications[{idx}].definition.language is missing English.")

    def check_aasc_3a_003(self):
        """
        Constraint AASc-3a-003: For a ConceptDescription referenced via ValueList/valueId and using data
        specification template IEC61360 (https://adminshell.io/DataSpecificationTemplates/DataSpecificationIec61360/3), DataSpecificationIec61360/value shall be
        set.
        -> not checked
        """
        pass

# 5.3.9 Environment


@dataclass
class Environment:
    asset_administration_shells: Optional[List[AssetAdministrationShell]]
    submodels: Optional[List[Submodel]]
    concept_descriptions: Optional[List[ConceptDescription]]



from unittest import TestCase
import subprocess
import os

script_dir = os.path.dirname(os.path.realpath(__file__))


class CheckFileCli(TestCase):

    json_file = os.path.join(script_dir, 'fixtures', 'aasx', 'valid', 'json', 'aasx', 'the_aas.json')
    xml_file = os.path.join(script_dir, 'fixtures', 'aasx', 'valid', 'xml', 'aasx', 'the_aas.xml')

    def invoke(self, args: list):
        result = subprocess.check_output(["python", "-m", "aas_test_engines", "check_file"] + args)
        return result.decode()

    def test_no_arguments(self):
        with self.assertRaises(subprocess.CalledProcessError):
            self.invoke([])

    def test_json(self):
        self.invoke([self.json_file, '--format', 'json'])

    def test_xml(self):
        self.invoke([self.xml_file, '--format', 'xml'])

    def test_html_output(self):
        result = self.invoke([self.json_file, '--format', 'json', '--output', 'html'])
        self.assertTrue(result.startswith('<!DOCTYPE html>'))

    def test_invalid_file(self):
        with self.assertRaises(subprocess.CalledProcessError):
            self.invoke([self.json_file, '--format', 'xml'])


class CheckServerCli(TestCase):

    def invoke(self, args: list):
        result = subprocess.check_output(["python", "-m", "aas_test_engines", "check_server"] + args)
        return result.decode()

    def test_no_arguments(self):
        with self.assertRaises(subprocess.CalledProcessError):
            self.invoke([])

    def test_suite_ambiguous(self):
        with self.assertRaises(subprocess.CalledProcessError):
            self.invoke(["https://localhost:5000", "RepositoryServiceSpecification/SSP-002", "--dry"])

    def test_dry(self):
        self.invoke(["https://localhost:5000", "https://admin-shell.io/aas/API/3/0/AssetAdministrationShellRepositoryServiceSpecification/SSP-002", "--dry"])



from typing import Dict, List, Tuple
from aas_test_engines.test_cases.v3_0.api import ExecConf
from aas_test_engines.test_cases.v3_0 import api as v3_0
from fences.core.util import ConfusionMatrix
from .result import AasTestResult
from .exception import AasTestToolsException

_DEFAULT_VERSION = '3.0'


def supported_versions() -> Dict[str, List[str]]:
    return {'3.0': v3_0._available_suites.keys()}


def latest_version():
    return _DEFAULT_VERSION


def execute_tests(conf: ExecConf, suite: str, version: str = _DEFAULT_VERSION) -> Tuple[AasTestResult, ConfusionMatrix]:
    if version != _DEFAULT_VERSION:
        raise AasTestToolsException(
            f"Unknown version {version}, must be one of {supported_versions()}")
    return v3_0.execute_tests(conf, suite)



import argparse
import sys
import os
import json
from aas_test_engines import api, file
from enum import Enum

# https://stackoverflow.com/questions/27981545
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

class InputFormats(Enum):
    xml = 'xml'
    json = 'json'
    aasx = 'aasx'

    def __str__(self):
        return self.value


class OutputFormats(Enum):
    TEXT = 'text'
    JSON = 'json'
    HTML = 'html'


def run_file_test(argv):
    parser = argparse.ArgumentParser(description='Checks a file for compliance with the AAS meta-model')
    parser.add_argument('file',
                        type=argparse.FileType('rb'),
                        help='the file to check')
    parser.add_argument('--format',
                        type=InputFormats,
                        default=InputFormats.aasx,
                        choices=list(InputFormats))
    parser.add_argument('--output',
                        type=OutputFormats,
                        default=OutputFormats.TEXT,
                        choices=list(OutputFormats))
    args = parser.parse_args(argv)

    if args.format == InputFormats.aasx:
        result = file.check_aasx_file(args.file)
    elif args.format == InputFormats.json:
        result = file.check_json_file(args.file)
    elif args.format == InputFormats.xml:
        result = file.check_xml_file(args.file)
    else:
        raise Exception(f"Invalid format {args.format}")
    if args.output == OutputFormats.TEXT:
        result.dump()
    elif args.output == OutputFormats.HTML:
        print(result.to_html())
    elif args.output == OutputFormats.JSON:
        print(json.dumps(result.to_dict()))
    else:
        raise Exception(f"Invalid output {args.output}")
    sys.exit(0 if result.ok() else 1)


def run_api_test(argv):
    parser = argparse.ArgumentParser(description='Checks a server instance for compliance with the AAS api')
    parser.add_argument('server',
                        type=str,
                        help='server to run the tests against')
    parser.add_argument('suite',
                        type=str,
                        help='test suite (or substring of it)')
    parser.add_argument('--dry',
                        action='store_true',
                        help="dry run, do not send requests")
    parser.add_argument('--version',
                        type=str,
                        default=api.latest_version())
    parser.add_argument('--no-verify',
                        action='store_true',
                        help='do not check TLS certificate')
    parser.add_argument('--remove-path-prefix',
                        type=str,
                        default='',
                        help='remove prefix from all paths')
    parser.add_argument('--output',
                        type=OutputFormats,
                        default=OutputFormats.TEXT,
                        choices=list(OutputFormats))
    args = parser.parse_args(argv)
    try:
        available_suites = api.supported_versions()[args.version]
    except KeyError:
        sys.stderr.write(f"Unknown version, must be one of {api.supported_versions().keys()}\n")
    suites = [i for i in available_suites if args.suite in i]
    if len(suites) == 0:
        sys.stderr.write(f"Unknown suite '{args.suite}', must be one of:\n")
        for i in available_suites:
            sys.stderr.writelines(f" - {i}\n")
        sys.exit(1)
    elif len(suites) > 1:
        sys.stderr.write(f"Substring '{args.suite}' is not unique, covers:\n")
        for i in suites:
            sys.stderr.write(f" - {i}\n")
        sys.exit(1)
    else:
        suite = suites[0]

    exec_conf = api.ExecConf(
        server=args.server,
        dry=args.dry,
        verify=not args.no_verify,
        remove_path_prefix=args.remove_path_prefix,
    )
    result, mat = api.execute_tests(exec_conf, suite, args.version)
    if args.output == OutputFormats.TEXT:
        result.dump()
    elif args.output == OutputFormats.HTML:
        print(result.to_html())
    elif args.output == OutputFormats.JSON:
        print(json.dumps(result.to_dict()))
    else:
        raise Exception(f"Invalid output {args.output}")
    sys.exit(0 if result.ok() else 1)


def generate_files(argv):
    parser = argparse.ArgumentParser(description='Generates aas files which can be used to test your software')
    parser.add_argument('directory',
                        type=str,
                        help='Directory to place files in')
    args = parser.parse_args(argv)
    if os.path.exists(args.directory):
        print(f"Directory '{args.directory}' already exists, please remove it")
        sys.exit(1)
    os.mkdir(args.directory)
    for idx, (is_valid, sample) in enumerate(file.generate()):
        tag = 'valid' if is_valid else 'invalid'
        with open(os.path.join(args.directory, f"{idx}_{tag}.json"), "w") as f:
            json.dump(sample, f)


commands = {
    'check_file': run_file_test,
    'check_server': run_api_test,
    'generate_files': generate_files,
}


def main():

    if len(sys.argv) <= 1:
        print(f"Usage: {sys.argv[0]} COMMAND OPTIONS...")
        print("Available commands:")
        print("  check_file      Check a file for compliance.")
        print("  check_server    Check a server instance for compliance.")
        print("  generate_files  Generate files for testing")
        sys.exit(1)

    command = sys.argv[1]

    if command not in commands:
        print(f"Unknown command '{command}', must be one of {', '.join(commands)}")
        sys.exit(1)

    remaining_args = sys.argv[2:]
    commands[command](remaining_args)


if __name__ == '__main__':
    main()



from decimal import Decimal, DecimalException
from typing import Pattern, Dict, Mapping
import re
import base64
from enum import Enum
from .parse import CheckConstraintException


def is_decimal(s: str) -> bool:
    try:
        Decimal(s)
        return True
    except DecimalException:
        return False


def _is_bounded_double(s: str, min: float, max: float) -> bool:
    if s in ['INF', '-INF', 'NaN']:
        return True
    try:
        v = float(s)
        return v > min and v < max
    except ValueError:
        return False


def is_float(s: str) -> bool:
    return _is_bounded_double(s, -3.4028234663852886e+38, 3.4028234663852886e+38)


def is_double(s: str) -> bool:
    return _is_bounded_double(s, -1.7976931348623158e+308, 1.7976931348623158e+308)


def _is_bounded_integer(s: int, min: int, max: int) -> bool:
    try:
        v = int(s)
        return v >= min and v <= max
    except ValueError:
        return False


def is_hex_binary(s: str) -> bool:
    try:
        bytes.fromhex(s)
        return True
    except (KeyError, ValueError):
        return False


def _is_leap_year(year: int) -> bool:
    if year < 0:
        year = abs(year) - 1
    if year % 4 > 0:
        return False
    if year % 100 > 0:
        return True
    if year % 400 > 0:
        return False
    return True


_DAYS_IN_MONTH: Mapping[int, int] = {
    1: 31,
    # Please use _is_leap_year if you need to check
    # whether a concrete February has 28 or 29 days.
    2: 29,
    3: 31,
    4: 30,
    5: 31,
    6: 30,
    7: 31,
    8: 31,
    9: 30,
    10: 31,
    11: 30,
    12: 31,
}


def _construct_matches_xs_date() -> Pattern[str]:
    digit = "[0-9]"
    year_frag = f"-?(([1-9]{digit}{digit}{digit}+)|(0{digit}{digit}{digit}))"
    month_frag = "((0[1-9])|(1[0-2]))"
    day_frag = f"((0[1-9])|([12]{digit})|(3[01]))"
    minute_frag = f"[0-5]{digit}"
    timezone_frag = f"(Z|(\\+|-)((0{digit}|1[0-3]):{minute_frag}|14:00))"
    date_lexical_rep = f"{year_frag}-{month_frag}-{day_frag}{timezone_frag}?"
    pattern = f"^{date_lexical_rep}$"

    return re.compile(pattern)


_REGEX_MATCHES_XS_DATE = _construct_matches_xs_date()

_DATE_PREFIX_RE = re.compile(r"^(-?[0-9]+)-([0-9]{2})-([0-9]{2})")


def is_xs_date(value: str) -> bool:
    if _REGEX_MATCHES_XS_DATE.match(value) is None:
        return False

    match = _DATE_PREFIX_RE.match(value)
    assert match is not None

    year = int(match.group(1))
    month = int(match.group(2))
    day = int(match.group(3))

    if year == 0:
        return False
    if day <= 0:
        return False
    if month <= 0 or month >= 13:
        return False

    if month == 2:
        max_days = 29 if _is_leap_year(year) else 28
    else:
        max_days = _DAYS_IN_MONTH[month]

    if day > max_days:
        return False

    return True


def _construct_matches_xs_date_time() -> Pattern[str]:
    digit = "[0-9]"
    year_frag = f"-?(([1-9]{digit}{digit}{digit}+)|(0{digit}{digit}{digit}))"
    month_frag = "((0[1-9])|(1[0-2]))"
    day_frag = f"((0[1-9])|([12]{digit})|(3[01]))"
    hour_frag = f"(([01]{digit})|(2[0-3]))"
    minute_frag = f"[0-5]{digit}"
    second_frag = f"([0-5]{digit})(\\.{digit}+)?"
    end_of_day_frag = "24:00:00(\\.0+)?"
    timezone_frag = f"(Z|(\\+|-)((0{digit}|1[0-3]):{minute_frag}|14:00))"
    date_time_lexical_rep = f"{year_frag}-{month_frag}-{day_frag}T(({hour_frag}:{minute_frag}:{second_frag})|{end_of_day_frag}){timezone_frag}?"
    pattern = f"^{date_time_lexical_rep}$"

    return re.compile(pattern)


_REGEX_MATCHES_XS_DATE_TIME = _construct_matches_xs_date_time()


def is_xs_date_time(s: str) -> bool:
    if _REGEX_MATCHES_XS_DATE_TIME.match(s) is None:
        return False

    date, _ = s.split("T")
    return is_xs_date(date)


def _construct_matches_xs_date_time_utc() -> Pattern[str]:
    digit = "[0-9]"
    year_frag = f"-?(([1-9]{digit}{digit}{digit}+)|(0{digit}{digit}{digit}))"
    month_frag = "((0[1-9])|(1[0-2]))"
    day_frag = f"((0[1-9])|([12]{digit})|(3[01]))"
    hour_frag = f"(([01]{digit})|(2[0-3]))"
    minute_frag = f"[0-5]{digit}"
    second_frag = f"([0-5]{digit})(\\.{digit}+)?"
    end_of_day_frag = "24:00:00(\\.0+)?"
    timezone_frag = "(Z|\\+00:00|-00:00)"
    date_time_lexical_rep = f"{year_frag}-{month_frag}-{day_frag}T(({hour_frag}:{minute_frag}:{second_frag})|{end_of_day_frag}){timezone_frag}"
    pattern = f"^{date_time_lexical_rep}$"

    return re.compile(pattern)


_REGEX_MATCHES_XS_DATE_TIME_UTC = _construct_matches_xs_date_time_utc()


def is_xs_date_time_utc(s: str) -> bool:
    if _REGEX_MATCHES_XS_DATE_TIME_UTC.match(s) is None:
        return False

    date, _ = s.split("T")
    return is_xs_date(date)


_DAYS_IN_MONTH: Dict[int, int] = {
    1: 31,
    2: 29,
    3: 31,
    4: 30,
    5: 31,
    6: 30,
    7: 31,
    8: 31,
    9: 30,
    10: 31,
    11: 30,
    12: 31,
}


def _days_in_month(month: int, year: int) -> int:
    if month == 2:
        return 29 if _is_leap_year(year) else 28
    else:
        return _DAYS_IN_MONTH[month]


def _construct_matches_xs_any_uri() -> Pattern[str]:
    scheme = "[a-zA-Z][a-zA-Z0-9+\\-.]*"
    ucschar = "[\\xa0-\\ud7ff\\uf900-\\ufdcf\\ufdf0-\\uffef\\U00010000-\\U0001fffd\\U00020000-\\U0002fffd\\U00030000-\\U0003fffd\\U00040000-\\U0004fffd\\U00050000-\\U0005fffd\\U00060000-\\U0006fffd\\U00070000-\\U0007fffd\\U00080000-\\U0008fffd\\U00090000-\\U0009fffd\\U000a0000-\\U000afffd\\U000b0000-\\U000bfffd\\U000c0000-\\U000cfffd\\U000d0000-\\U000dfffd\\U000e1000-\\U000efffd]"
    iunreserved = f"([a-zA-Z0-9\\-._~]|{ucschar})"
    pct_encoded = "%[0-9A-Fa-f][0-9A-Fa-f]"
    sub_delims = "[!$&'()*+,;=]"
    iuserinfo = f"({iunreserved}|{pct_encoded}|{sub_delims}|:)*"
    h16 = "[0-9A-Fa-f]{1,4}"
    dec_octet = "([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])"
    ipv4address = f"{dec_octet}\\.{dec_octet}\\.{dec_octet}\\.{dec_octet}"
    ls32 = f"({h16}:{h16}|{ipv4address})"
    ipv6address = f"(({h16}:){{6}}{ls32}|::({h16}:){{5}}{ls32}|({h16})?::({h16}:){{4}}{ls32}|(({h16}:)?{h16})?::({h16}:){{3}}{ls32}|(({h16}:){{2}}{h16})?::({h16}:){{2}}{ls32}|(({h16}:){{3}}{h16})?::{h16}:{ls32}|(({h16}:){{4}}{h16})?::{ls32}|(({h16}:){{5}}{h16})?::{h16}|(({h16}:){{6}}{h16})?::)"
    unreserved = "[a-zA-Z0-9\\-._~]"
    ipvfuture = f"[vV][0-9A-Fa-f]+\\.({unreserved}|{sub_delims}|:)+"
    ip_literal = f"\\[({ipv6address}|{ipvfuture})\\]"
    ireg_name = f"({iunreserved}|{pct_encoded}|{sub_delims})*"
    ihost = f"({ip_literal}|{ipv4address}|{ireg_name})"
    port = "[0-9]*"
    iauthority = f"({iuserinfo}@)?{ihost}(:{port})?"
    ipchar = f"({iunreserved}|{pct_encoded}|{sub_delims}|[:@])"
    isegment = f"({ipchar})*"
    ipath_abempty = f"(/{isegment})*"
    isegment_nz = f"({ipchar})+"
    ipath_absolute = f"/({isegment_nz}(/{isegment})*)?"
    ipath_rootless = f"{isegment_nz}(/{isegment})*"
    ipath_empty = f"({ipchar}){{0}}"
    ihier_part = f"(//{iauthority}{ipath_abempty}|{ipath_absolute}|{ipath_rootless}|{ipath_empty})"
    iprivate = "[\\ue000-\\uf8ff\\U000f0000-\\U000ffffd\\U00100000-\\U0010fffd]"
    iquery = f"({ipchar}|{iprivate}|[/?])*"
    ifragment = f"({ipchar}|[/?])*"
    isegment_nz_nc = f"({iunreserved}|{pct_encoded}|{sub_delims}|@)+"
    ipath_noscheme = f"{isegment_nz_nc}(/{isegment})*"
    irelative_part = f"(//{iauthority}{ipath_abempty}|{ipath_absolute}|{ipath_noscheme}|{ipath_empty})"
    irelative_ref = f"{irelative_part}(\\?{iquery})?(#{ifragment})?"
    iri = f"{scheme}:{ihier_part}(\\?{iquery})?(#{ifragment})?"
    iri_reference = f"({iri}|{irelative_ref})"
    pattern = f"^{iri_reference}$"

    return re.compile(pattern)


def _construct_matches_xs_g_month_day() -> Pattern[str]:
    g_month_day_rep = "--(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])(Z|(\\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))?"
    pattern = f"^{g_month_day_rep}$"

    return re.compile(pattern)


_REGEX_MATCHES_XS_G_MONTH_DAY = _construct_matches_xs_g_month_day()


def is_xs_g_month_day(s: str) -> bool:
    if _REGEX_MATCHES_XS_G_MONTH_DAY.match(s) is None:
        return False
    month = int(s[2:4])
    day = int(s[5:7])

    max_days = _DAYS_IN_MONTH[month]
    return day <= max_days


_REGEX_MATCHES_XS_ANY_URI = _construct_matches_xs_any_uri()


def is_any_uri(s: str) -> bool:
    return _REGEX_MATCHES_XS_ANY_URI.match(s) is not None


def is_base64_binary(s: str) -> bool:
    try:
        base64.b64decode(s)
        return True
    except ValueError:
        return False


def _construct_matches_bcp_47() -> Pattern[str]:
    alphanum = '[a-zA-Z0-9]'
    singleton = '[0-9A-WY-Za-wy-z]'
    extension = f'{singleton}(-({alphanum}){{2,8}})+'
    extlang = '[a-zA-Z]{3}(-[a-zA-Z]{3}){0,2}'
    irregular = '(en-GB-oed|i-ami|i-bnn|i-default|i-enochian|i-hak|i-klingon|i-lux|i-mingo|i-navajo|i-pwn|i-tao|i-tay|i-tsu|sgn-BE-FR|sgn-BE-NL|sgn-CH-DE)'
    regular = '(art-lojban|cel-gaulish|no-bok|no-nyn|zh-guoyu|zh-hakka|zh-min|zh-min-nan|zh-xiang)'
    grandfathered = f'({irregular}|{regular})'
    language = f'([a-zA-Z]{{2,3}}(-{extlang})?|[a-zA-Z]{{4}}|[a-zA-Z]{{5,8}})'
    script = '[a-zA-Z]{4}'
    region = '([a-zA-Z]{2}|[0-9]{3})'
    variant = f'(({alphanum}){{5,8}}|[0-9]({alphanum}){{3}})'
    privateuse = f'[xX](-({alphanum}){{1,8}})+'
    langtag = f'{language}(-{script})?(-{region})?(-{variant})*(-{extension})*(-{privateuse})?'
    language_tag = f'({langtag}|{privateuse}|{grandfathered})'
    pattern = f'^{language_tag}$'

    return re.compile(pattern)


_REGEX_MATCHES_BCP_47 = _construct_matches_bcp_47()


def is_bcp_lang_string(s: str) -> bool:
    return _REGEX_MATCHES_BCP_47.match(s) is not None


def _construct_matches_version_type() -> Pattern[str]:
    pattern = '^(0|[1-9][0-9]*)$'
    return re.compile(pattern)


_REGEX_MATCHES_VERSION_TYPE = _construct_matches_version_type()


def is_version_string(text: str) -> bool:
    return _REGEX_MATCHES_VERSION_TYPE.match(text) is not None


_REGEX_IS_BCP_47_FOR_ENGLISH = re.compile("^(en|EN)(-.*)?$")


def is_bcp_47_for_english(text: str) -> bool:
    return _REGEX_IS_BCP_47_FOR_ENGLISH.match(text) is not None


class DataTypeDefXsd(Enum):
    anyURI = "xs:anyURI"
    base64Binary = "xs:base64Binary"
    boolean = "xs:boolean"
    byte = "xs:byte"
    date = "xs:date"
    dateTime = "xs:dateTime"
    decimal = "xs:decimal"
    double = "xs:double"
    duration = "xs:duration"
    float = "xs:float"
    gDay = "xs:gDay"
    gMonth = "xs:gMonth"
    gMonthDay = "xs:gMonthDay"
    gYear = "xs:gYear"
    gYearMonth = "xs:gYearMonth"
    hexBinary = "xs:hexBinary"
    int = "xs:int"
    integer = "xs:integer"
    long = "xs:long"
    negativeInteger = "xs:negativeInteger"
    nonNegativeInteger = "xs:nonNegativeInteger"
    nonPositiveInteger = "xs:nonPositiveInteger"
    positiveInteger = "xs:positiveInteger"
    short = "xs:short"
    string = "xs:string"
    time = "xs:time"
    unsignedByte = "xs:unsignedByte"
    unsignedInt = "xs:unsignedInt"
    unsignedLong = "xs:unsignedLong"
    unsignedShort = "xs:unsignedShort"


validators = {
    DataTypeDefXsd.string: lambda _: True,
    DataTypeDefXsd.boolean: lambda x: x in {"true", "false", "1", "0"},
    DataTypeDefXsd.decimal: is_decimal,
    DataTypeDefXsd.integer: lambda s: _is_bounded_integer(s, float('-inf'), float('inf')),

    DataTypeDefXsd.float: is_float,
    DataTypeDefXsd.double: is_double,

    DataTypeDefXsd.byte: lambda s: _is_bounded_integer(s, -128, 127),
    DataTypeDefXsd.short: lambda s: _is_bounded_integer(s, -32768, 32767),
    DataTypeDefXsd.int: lambda s: _is_bounded_integer(s, -2147483648, 2147483647),
    DataTypeDefXsd.long: lambda s: _is_bounded_integer(s, -9223372036854775808, 9223372036854775807),

    DataTypeDefXsd.unsignedByte: lambda s: _is_bounded_integer(s, 0, 255),
    DataTypeDefXsd.unsignedShort: lambda s: _is_bounded_integer(s, 0, 65535),
    DataTypeDefXsd.unsignedInt: lambda s: _is_bounded_integer(s, 0, 4294967295),
    DataTypeDefXsd.unsignedLong: lambda s: _is_bounded_integer(s, 0, 18446744073709551615),

    DataTypeDefXsd.positiveInteger: lambda s: _is_bounded_integer(s, 1, float('inf')),
    DataTypeDefXsd.nonNegativeInteger: lambda s: _is_bounded_integer(s, 0, float('inf')),
    DataTypeDefXsd.negativeInteger: lambda s: _is_bounded_integer(s, float('-inf'), -1),
    DataTypeDefXsd.nonPositiveInteger: lambda s: _is_bounded_integer(s, float('-inf'), 0),

    DataTypeDefXsd.date: is_xs_date,
    DataTypeDefXsd.dateTime: is_xs_date_time,
    DataTypeDefXsd.gMonthDay: is_xs_g_month_day,
    DataTypeDefXsd.gYearMonth: lambda x: re.fullmatch(r"-?([1-9][0-9]{3,}|0[0-9]{3})-(0[1-9]|1[0-2])(Z|(\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))?", x) is not None,
    DataTypeDefXsd.gDay: lambda x: re.fullmatch(r"---(0[1-9]|[12][0-9]|3[01])(Z|(\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))?", x) is not None,
    DataTypeDefXsd.gMonth: lambda x: re.fullmatch(r"--(0[1-9]|1[0-2])(Z|(\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))?", x) is not None,
    DataTypeDefXsd.gYear: lambda x: re.fullmatch(r"-?([1-9][0-9]{3,}|0[0-9]{3})(Z|(\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))?", x) is not None,
    DataTypeDefXsd.time: lambda x: re.fullmatch(r"(([01][0-9]|2[0-3]):[0-5][0-9]:[0-5][0-9](\.[0-9]+)?|(24:00:00(\.0+)?))(Z|(\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))?", x) is not None,
    DataTypeDefXsd.duration: lambda x: re.fullmatch(r"-?P((([0-9]+Y([0-9]+M)?([0-9]+D)?|([0-9]+M)([0-9]+D)?|([0-9]+D))(T(([0-9]+H)([0-9]+M)?([0-9]+(\.[0-9]+)?S)?|([0-9]+M)([0-9]+(\.[0-9]+)?S)?|([0-9]+(\.[0-9]+)?S)))?)|(T(([0-9]+H)([0-9]+M)?([0-9]+(\.[0-9]+)?S)?|([0-9]+M)([0-9]+(\.[0-9]+)?S)?|([0-9]+(\.[0-9]+)?S))))", x) is not None,

    DataTypeDefXsd.anyURI: is_any_uri,
    DataTypeDefXsd.base64Binary: is_base64_binary,
    DataTypeDefXsd.hexBinary: is_hex_binary,
}



from unittest import TestCase
from aas_test_engines._util import group, un_group, normpath, splitpath


class GroupSmcTest(TestCase):

    def test_empty(self):
        result = group({})
        self.assertDictEqual(result, {})
        result = un_group({})
        self.assertDictEqual(result, {})

    def test_simple(self):
        input = {
            'modelType': 'SubmodelElementCollection',
            'value': [
                {
                    'idShort': 'x'
                }
            ]
        }
        result = group(input)
        self.assertDictEqual(result, {
            'modelType': 'SubmodelElementCollection',
            'value': {
                'x': [{'idShort': 'x'}]
            }
        })
        output = un_group(result)
        self.assertDictEqual(input, output)

    def test_multiple(self):
        input = {
            'modelType': 'SubmodelElementCollection',
            'value': [
                {
                    'idShort': 'x',
                    'x': 1
                },
                {
                    'idShort': 'z',
                    'z': 2
                },
                {
                    'idShort': 'x',
                    'x': 3
                }
            ]
        }
        result = group(input)
        self.assertDictEqual(result, {
            'modelType': 'SubmodelElementCollection',
            'value': {
                'x': [{'idShort': 'x', 'x': 1}, {'idShort': 'x', 'x': 3}],
                'z': [{'idShort': 'z', 'z': 2}]
            }
        })
        output = un_group(result)
        self.assertEqual(output['modelType'], 'SubmodelElementCollection')
        self.assertEqual(len(output['value']), 3)

    def test_nested(self):
        input = {
            'modelType': 'SubmodelElementCollection',
            'value': [
                {
                    'modelType': 'SubmodelElementCollection',
                    'idShort': 'foo',
                    'value': [
                        {
                            'idShort': 'x'
                        }
                    ]
                }
            ]
        }
        result = group(input)


class NormPathTest(TestCase):

    def test_empty(self):
        self.assertEqual(normpath(''), '')
        self.assertEqual(normpath('   '), '')

    def test_double_slashes(self):
        self.assertEqual(normpath('//a//b//c'), '/a/b/c')
        self.assertEqual(normpath('//'), '/')

    def test_dot(self):
        self.assertEqual(normpath('.'), '')
        self.assertEqual(normpath('a/./b/./c'), 'a/b/c')
        self.assertEqual(normpath('a/b/.'), 'a/b')

    def test_trailing_slashes(self):
        self.assertEqual(normpath('/abc/def/'), '/abc/def')
        self.assertEqual(normpath('abc/def// //'), 'abc/def')

    def test_double_dot(self):
        self.assertEqual(normpath('foo/../bar'), 'bar')
        self.assertEqual(normpath('/foo/../bar///'), '/bar')
        self.assertEqual(normpath('a/b/../c'), 'a/c')
        self.assertEqual(normpath('a/b/../c/..'), 'a')
        self.assertEqual(normpath('/a/b/../c/..'), '/a')

    def test_illegal_double_dot(self):
        self.assertEqual(normpath('..'), '')
        self.assertEqual(normpath('/..'), '/')
        self.assertEqual(normpath('../x'), 'x')
        self.assertEqual(normpath('/../x'), '/x')


class SplitPathTest(TestCase):

    def test_empty(self):
        self.assertEqual(splitpath(''), ('', ''))

    def test_ends_with_slash(self):
        self.assertEqual(splitpath('a/b.txt/'), ('a/b.txt', ''))
        self.assertEqual(splitpath('/'), ('', ''))
        self.assertEqual(splitpath('/a/b/'), ('/a/b', ''))

    def test_no_slash(self):
        self.assertEqual(splitpath('foo.txt'), ('', 'foo.txt'))



from .adapter import Adapter, AdapterException

from dataclasses import dataclass, fields, field, is_dataclass
from typing import List, Dict, Optional, Tuple, Union, ForwardRef, Pattern, Callable
from aas_test_engines.result import AasTestResult, Level
from enum import Enum
import re
from .adapter import AdapterPath, JsonAdapter, XmlAdapter


class CheckConstraintException(Exception):
    pass


INVALID = object()


def to_lower_camel_case(snake_str):
    camel_string = "".join(x.capitalize() for x in snake_str.lower().split("_"))
    return snake_str[0].lower() + camel_string[1:]


def unwrap_optional(cls) -> Tuple[bool, type]:
    try:
        origin = cls.__origin__
    except AttributeError:
        return True, cls
    if origin is not Union:
        return True, cls
    # We only support Optional aka. Union[x, NoneType]
    assert len(cls.__args__) == 2 and cls.__args__[1] is type(None)
    return False, cls.__args__[0]


def abstract(cls):
    # Prefix with cls to apply only to base class
    setattr(cls, f'_{cls}_abstract', True)
    return cls


def isabstract(cls):
    return hasattr(cls, f'_{cls}_abstract')


def requires_model_type(cls):
    setattr(cls, f'_requires_model_type', True)
    return cls


def has_requires_model_type(cls):
    return hasattr(cls, f'_requires_model_type')


def collect_subclasses(cls, result: Dict[str, type]):
    if not isabstract(cls):
        result[cls.__name__] = cls
    for i in cls.__subclasses__():
        collect_subclasses(i, result)


class StringFormattedValue:
    min_length: Optional[int] = None
    max_length: Optional[int] = None
    pattern: Optional[Pattern] = None

    def __init__(self, raw_value: str):
        self.raw_value = raw_value
        if self.min_length is not None:
            if len(raw_value) < self.min_length:
                raise ValueError(f"String is shorter than {self.min_length} characters")
        if self.max_length is not None:
            if len(raw_value) > self.max_length:
                raise ValueError(f"String is longer than {self.max_length} characters")

        # Constraint AASd-130: An attribute with data type "string" shall be restricted to the characters as defined in
        # XML Schema 1.0, i.e. the string shall consist of these characters only: ^[\x09\x0A\x0D\x20-\uD7FF\uE000-
        # \uFFFD\u00010000-\u0010FFFF]*$.
        if re.fullmatch(r"[\x09\x0a\x0d\x20-\ud7ff\ue000-\ufffd\U00010000-\U0010ffff]*", raw_value) is None:
            raise ValueError("Constraint AASd-130 violated: String is not XML serializable")

        if self.pattern:
            if re.fullmatch(self.pattern, raw_value) is None:
                raise ValueError(f"String '{raw_value}' does not match pattern {self.pattern}")

    def __eq__(self, other: "StringFormattedValue") -> bool:
        return self.raw_value == other.raw_value

    def __str__(self) -> str:
        return self.raw_value


def parse_string_formatted_value(cls, value: Adapter, result: AasTestResult) -> StringFormattedValue:
    try:
        return cls(value.as_string())
    except (AdapterException, ValueError) as e:
        result.append(AasTestResult(f"{e} @ {value.path}", level=Level.ERROR))
    return None


def parse_list(item_cls, value: Adapter, result: AasTestResult) -> list:
    try:
        items = value.as_list()
    except AdapterException as e:
        result.append(AasTestResult(f"{e} @ {value.path}", level=Level.ERROR))
        return INVALID
    return [parse(item_cls, i, result) for i in items]


def parse_bool(value: Adapter, result: AasTestResult) -> bool:
    try:
        return value.as_bool()
    except AdapterException as e:
        result.append(AasTestResult(f"{e} @ {value.path}", level=Level.ERROR))
        return INVALID


def parse_string(value: Adapter, result: AasTestResult) -> str:
    try:
        return value.as_string()
    except AdapterException as e:
        result.append(AasTestResult(f"{e} @ {value.path}", level=Level.ERROR))
        return INVALID


def parse_enum(cls, value: Adapter, result: AasTestResult):
    try:
        str_val = value.as_string()
    except AdapterException as e:
        result.append(AasTestResult(f"{e} @ {value.path}", level=Level.ERROR))
        return INVALID
    try:
        return cls(str_val)
    except ValueError as e:
        result.append(AasTestResult(f"{e} @ {value.path}", level=Level.ERROR))
    return INVALID


def parse_abstract_object(cls, adapter: Adapter, result: AasTestResult):
    try:
        discriminator = adapter.get_model_type()
    except AdapterException as e:
        result.append(AasTestResult(f"{e} @ {adapter.path}", level=Level.ERROR))
        return INVALID
    subclasses = {}
    collect_subclasses(cls, subclasses)
    try:
        cls = subclasses[discriminator]
    except KeyError:
        result.append(AasTestResult(f"Invalid model type {discriminator} @ {adapter.path}", level=Level.ERROR))
        return INVALID
    return parse_concrete_object(cls, adapter, result)


def parse_concrete_object(cls, adapter: Adapter, result: AasTestResult):
    try:
        obj = adapter.as_object()
    except AdapterException as e:
        result.append(AasTestResult(f"{e} @ {adapter.path}", level=Level.ERROR))
        return INVALID
    if has_requires_model_type(cls):
        try:
            discriminator = adapter.get_model_type()
            if discriminator != cls.__name__:
                result.append(AasTestResult(f"Wrong model type @ {adapter.path}", level=Level.ERROR))
        except AdapterException as e:
            result.append(AasTestResult(f"Model typ missing @ {adapter.path}", level=Level.ERROR))

    args = {}
    all_fields = set()
    for field in fields(cls):
        try:
            exclude_as = field.metadata['exclude_as']
            args[field.name] = exclude_as
            continue
        except KeyError:
            pass
        field_name = to_lower_camel_case(field.name)
        all_fields.add(field_name)
        required, field_type = unwrap_optional(field.type)
        try:
            obj_value = obj[field_name]
        except KeyError:
            if required:
                result.append(AasTestResult(f"Missing attribute {field_name} @ {adapter.path}", level=Level.ERROR))
                args[field.name] = INVALID
            else:
                args[field.name] = None
            continue
        args[field.name] = parse(field_type, obj_value, result)

    # Check unknown additional attributes
    for key in obj.keys():
        if key not in all_fields:
            result.append(AasTestResult(f"Unknown additional attribute {key} @ {adapter.path}", level=Level.ERROR))

    return cls(**args)


def parse(cls, obj_value: Adapter, result: AasTestResult):
    # Unwrap a forward reference
    if isinstance(cls, ForwardRef):
        # TODO: this is a hack to get our only ForwardRef["Reference"] into scope
        from .model import Reference
        try:
            cls = cls._evaluate(globals(), locals(), recursive_guard=frozenset())
        except TypeError:
            # Python < 3.9
            cls = cls._evaluate(globals(), locals())

    origin = getattr(cls, '__origin__', None)
    if origin:
        if origin is list:
            item_type = cls.__args__[0]
            return parse_list(item_type, obj_value, result)
    else:
        if is_dataclass(cls):
            if isabstract(cls):
                return parse_abstract_object(cls, obj_value, result)
            else:
                obj = parse_concrete_object(cls, obj_value, result)
                post_parse = getattr(obj, "post_parse", None)
                if post_parse and result.ok():
                    post_parse()
                return obj
        elif cls is bool:
            return parse_bool(obj_value, result)
        elif cls is str:
            return parse_string(obj_value, result)
        elif isinstance(cls, Enum.__class__):
            return parse_enum(cls, obj_value, result)
        elif isinstance(cls, StringFormattedValue.__class__):
            return parse_string_formatted_value(cls, obj_value, result)
    raise NotImplementedError(
        f"There is no parsing implemented for:\n"
        f"  origin:    {origin}\n"
        f"  args:      {getattr(cls, '__args__', None)}\n"
        f"  obj_value: {obj_value}\n"
        f"  cls:       {cls}\n"
    )


def check_constraints(obj, result: AasTestResult, path: AdapterPath = AdapterPath()):
    if not is_dataclass(obj):
        return
    fns = [getattr(obj, i) for i in dir(obj) if i.startswith('check_')]
    for fn in fns:
        try:
            fn()
        except CheckConstraintException as e:
            result.append(AasTestResult(f"{e} @ {path}", level=Level.ERROR))
    for field in fields(obj):
        value = getattr(obj, field.name)
        if isinstance(value, list):
            for idx, i in enumerate(value):
                check_constraints(i, result, path + field.name + idx)
        else:
            check_constraints(value, result, path + field.name)


def _parse_and_check(cls, adapter: Adapter) -> Tuple[object, AasTestResult]:
    result_root = AasTestResult("Check")
    result_meta_model = AasTestResult("Check meta model")
    env = parse(cls, adapter, result_meta_model)
    result_root.append(result_meta_model)
    if result_root.ok():
        result_constraints = AasTestResult("Check constraints")
        check_constraints(env, result_constraints)
        result_root.append(result_constraints)
    else:
        result_root.append(AasTestResult("Skipped checking of constraints", Level.WARNING))
    return result_root, env


def parse_and_check_json(cls, value: any) -> Tuple[AasTestResult, object]:
    return _parse_and_check(cls, JsonAdapter(value, AdapterPath()))


def parse_and_check_xml(cls, value: any) -> Tuple[AasTestResult, object]:
    return _parse_and_check(cls, XmlAdapter(value, AdapterPath()))



from typing import List, TypeVar, Union
from enum import Enum
import os
import html

T = TypeVar("T")


script_dir = os.path.dirname(os.path.realpath(__file__))


class Level(Enum):
    INFO = 0
    WARNING = 1
    ERROR = 2
    CRITICAL = 3

    def __or__(self, other: "Level") -> "Level":
        return Level(max(self.value, other.value))

    def color(self) -> str:
        if self.value == 0:
            return '\033[92m'
        if self.value == 1:
            return '\033[93m'
        if self.value == 2:
            return '\033[91m'
        return '\033[94m'


class AasTestResult:

    def __init__(self, message: str, level=Level.INFO):
        assert isinstance(level, Level)
        self.message = message
        self.level = level
        self.sub_results: List[AasTestResult] = []

    def append(self, result: "AasTestResult"):
        self.sub_results.append(result)
        self.level = self.level | result.level

    def ok(self) -> bool:
        return self.level == Level.INFO or self.level == Level.WARNING

    def dump(self):
        """Outputs the result to console"""
        for line in self.to_lines():
            print(line)

    def to_lines(self, indent=0, path=''):
        ENDC = '\033[0m'
        yield "   " * indent + self.level.color() + self.message + ENDC
        for sub_result in self.sub_results:
            yield from sub_result.to_lines(indent + 1)

    def _to_html(self, level: int) -> str:
        cls = {
            Level.INFO: 'info',
            Level.WARNING: 'warning',
            Level.ERROR: 'error',
            Level.CRITICAL: 'critical',
        }[self.level]
        s = "<div>\n"
        msg = html.escape(self.message)
        if self.sub_results:
            c = "" if self.ok() else "caret-down"
            s += f'<div class="{cls}">{msg}<span class="caret level-{level} {c}"/></div>\n'
            c = "" if self.ok() else "visible"
            s += f'<div class="sub-results {c}">\n'
            for sub_result in self.sub_results:
                s += sub_result._to_html(level + 1)
            s += "</div>\n"
        else:
            s += f'<div class="{cls}">{msg}</div>\n'
        s += "</div>\n"
        return s

    def to_html(self) -> str:
        """Generates an interactive view of the result as HTML.
        You should write the result into a file:
            content = self.result.to_html()
            with open("result.html", "w") as file:
                file.write(content)
        Then open result.html in your browser.
        """
        with open(os.path.join(script_dir, 'data', 'template.html'), 'r') as f:
            content = f.read()
        return content.replace("<!-- CONTENT -->", self._to_html(0))

    def to_dict(self):
        return {
            'm': self.message,
            'l': self.level.value,
            's': [i.to_dict() for i in self.sub_results]
        }

    @classmethod
    def from_json(self, data: dict) -> "AasTestResult":
        v = AasTestResult(
            data['m'], Level(data['l'])
        )
        for i in data['s']:
            v.append(AasTestResult.from_json(i))
        return v


managers: List["ContextManager"] = []


class ContextManager:

    def __init__(self, result: AasTestResult, catch_all_exceptions: bool):
        self.result = result
        self.catch_all_exceptions = catch_all_exceptions

    def __enter__(self) -> AasTestResult:
        managers.append(self)
        return self.result

    def __exit__(self, exc_type, exc_val, traceback):
        m = managers.pop()
        assert m is self
        if exc_val is None:
            if managers:
                managers[-1].result.append(self.result)
        elif isinstance(exc_val, ResultException):
            self.result.append(exc_val.result)
            if managers:
                managers[-1].result.append(self.result)
            return True
        elif self.catch_all_exceptions:
            self.result.append(AasTestResult(f"Internal error: {exc_val}", Level.CRITICAL))
            if managers:
                managers[-1].result.append(self.result)
            return True
        else:
            return False


class ResultException(Exception):
    def __init__(self, result) -> None:
        self.result = result


def _as_result(message: Union[str, AasTestResult], level: Level):
    if isinstance(message, AasTestResult):
        return message
    assert isinstance(message, str)
    return AasTestResult(message, level=level)


def write(message: Union[str, AasTestResult]):
    if not managers:
        raise RuntimeError("No open context")
    message = _as_result(message, Level.INFO)
    managers[-1].result.append(message)


def start(message: Union[str, AasTestResult], catch_all_exceptions: bool = False) -> ContextManager:
    result = _as_result(message, Level.INFO)
    return ContextManager(result, catch_all_exceptions)


def abort(message: Union[str, AasTestResult]):
    result = _as_result(message, Level.ERROR)
    raise ResultException(result)



from aas_test_engines.result import AasTestResult, Level, write, start, abort, ResultException
from html.parser import HTMLParser

from unittest import TestCase


class ResultTest(TestCase):

    def setUp(self) -> None:
        self.result = AasTestResult('test', Level.INFO)
        self.result.append(AasTestResult('test1', Level.INFO))
        sub_result = AasTestResult('test2')
        self.result.append(sub_result)
        self.result.append(AasTestResult('test3', Level.ERROR))
        sub_result.append(AasTestResult('sub', Level.WARNING))

    def test_ok(self):
        self.assertTrue(AasTestResult('', Level.INFO).ok())
        self.assertTrue(AasTestResult('', Level.WARNING).ok())
        self.assertFalse(AasTestResult('', Level.ERROR).ok())

    def test_append(self):
        result = AasTestResult('test', Level.INFO)
        self.assertEqual(result.level, Level.INFO)
        result.append(AasTestResult('test1', Level.INFO))
        self.assertEqual(result.level, Level.INFO)
        result.append(AasTestResult('test2', Level.WARNING))
        self.assertEqual(result.level, Level.WARNING)
        result.append(AasTestResult('test2',  Level.ERROR))
        self.assertEqual(result.level, Level.ERROR)
        self.assertEqual(len(result.sub_results), 3)

    def test_dump(self):
        self.result.dump()

    def test_to_json(self):
        j = self.result.to_dict()
        result = AasTestResult.from_json(j)
        self.assertEqual(len(self.result.sub_results), len(result.sub_results))
        self.assertEqual(self.result.level, result.level)
        self.assertEqual(self.result.message, result.message)

    def test_to_html(self):
        class SimpleHtmlChecker(HTMLParser):
            tags = []

            def handle_starttag(self, tag, attrs) -> None:
                self.tags.append(tag)

            def handle_endtag(self, tag) -> None:
                assert tag == self.tags.pop(-1)
        content = self.result.to_html()
        self.assertIn(self.result.message, content)
        checker = SimpleHtmlChecker()
        checker.feed(content)
        self.assertTrue(len(checker.tags) == 0)
        checker.close()


class ContextManagerTest(TestCase):

    def test_write_without_context(self):
        with self.assertRaises(RuntimeError):
            write("foo")

    def test_write_with_context(self):
        with start("foo") as r:
            write("bar")
        self.assertEqual(r.message, "foo")
        self.assertEqual(len(r.sub_results), 1)
        self.assertEqual(r.sub_results[0].message, "bar")
        self.assertEqual(len(r.sub_results[0].sub_results), 0)
        self.assertTrue(r.ok())

    def test_abort_without_context(self):
        with self.assertRaises(ResultException):
            abort("foo")

    def test_abort_with_context(self):
        with start("foo") as r:
            abort("bar")
        self.assertEqual(r.message, "foo")
        self.assertEqual(len(r.sub_results), 1)
        self.assertEqual(r.sub_results[0].message, "bar")
        self.assertEqual(len(r.sub_results[0].sub_results), 0)
        self.assertFalse(r.ok())
        r.dump()

    def test_start(self):
        with start('foo') as r:
            pass
        self.assertEqual(r.message, "foo")
        self.assertEqual(len(r.sub_results), 0)
        self.assertTrue(r.ok())
        r.dump()

    def test_nested_start(self):
        with start('foo') as r:
            write('foo_start')
            with start('bar'):
                write('bar_x')
            write('foo_end')

        self.assertEqual(r.message, "foo")
        self.assertEqual(len(r.sub_results), 3)
        self.assertEqual(r.sub_results[0].message, "foo_start")
        self.assertEqual(r.sub_results[1].message, "bar")
        self.assertEqual(r.sub_results[2].message, "foo_end")
        self.assertEqual(len(r.sub_results[1].sub_results), 1)
        self.assertEqual(r.sub_results[1].sub_results[0].message, "bar_x")
        self.assertEqual(len(r.sub_results[1].sub_results[0].sub_results), 0)
        self.assertTrue(r.ok())
        r.dump()

    def test_continue_after_abort(self):
        with start('foo') as r:
            write('foo_start')
            with start('bar'):
                abort('bar_x')
                write('bar_y')  # won't be reached
            write('foo_end')  # must be written

        r.dump()
        self.assertEqual(r.message, "foo")
        self.assertEqual(len(r.sub_results), 3)
        self.assertEqual(r.sub_results[0].message, "foo_start")
        self.assertEqual(r.sub_results[1].message, "bar")
        self.assertEqual(r.sub_results[2].message, "foo_end")
        self.assertEqual(len(r.sub_results[1].sub_results), 1)
        self.assertEqual(r.sub_results[1].sub_results[0].message, "bar_x")
        self.assertEqual(len(r.sub_results[1].sub_results[0].sub_results), 0)
        self.assertFalse(r.ok())

    def test_nested_abort(self):
        with start("foo") as r:
            with start("bar"):
                abort("x")
            abort("y")
        r.dump()
        self.assertEqual(r.message, "foo")
        self.assertEqual(len(r.sub_results), 2)
        self.assertEqual(r.sub_results[0].message, "bar")
        self.assertEqual(len(r.sub_results[0].sub_results), 1)
        self.assertEqual(r.sub_results[0].sub_results[0].message, 'x')
        self.assertEqual(r.sub_results[1].message, 'y')
        self.assertFalse(r.ok())

    def test_foreign_exception(self):
        with self.assertRaises(RuntimeError):
            with start("foo") as r:
                write("bar")
                raise RuntimeError("xyz")
        self.assertEqual(r.message, "foo")
        self.assertEqual(r.sub_results[0].message, "bar")
